/**
 * Engine.cpp:
 * 	a SDL-specific implementation of Engine.h
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Engine.h"

#include "platform/Framework.h"
#include "blurt/Raytracer.h"
#include "blurt/Scene.h"
#include "blurt/Ray.h"
#include "blurt/EngineError.h"
#include "util/CmdLineParser.h"
#include "util/Clock.h"
#include "defaultsettings.h"
#include <iostream>
#include <boost/scoped_ptr.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#ifdef _OPENMP
#include <omp.h>
#endif


using namespace std;
using namespace blurt;
using namespace platform;
using namespace util;

Engine::Engine()
	: isRunning(false),
	  doRender(true)
{
}


Engine::~Engine()
{
	tearDown();
}

Engine& Engine::get()
{
	static Engine e;
	return e;
}


bool Engine::setup(int argc, char** argv)
{
	if (!parseCommandLine(argc, argv))
		return false;
	platform::init();
	srand(time(0));
	scene.reset(new Scene());
	scene->readFromFile(inputfile);

	screen.reset(new platform::Screen(screenWidth, screenHeight, TITLE));
	raytracer.reset(new Raytracer(*screen.get(), recursionDepth));
	return true;
}


void Engine::run(int argc, char** argv)
{
	if (!setup(argc, argv))
		return;

	isRunning = true;

	Clock clock = Clock();

	if (!videofile.empty())
		raytracer->enableRenderToVideo(videofile.c_str());
	while (isRunning)
	{
		if (doRender)
		{
			doRender = !renderOnlyOneFrame;
			scene->update();
			Ray::resetRayCount();
			clock.startFrame();
			if (supersampling > 1)
				raytracer->renderSupersampled(supersampling);
			else
				raytracer->render();
			clock.endFrame();

			// display fps every 15 frames
			if (showFps && (clock.getFrameCount() & 0xf) == 0xf)
				cout << "FPS: " << clock.getFps() << '\n';

			unsigned r = Ray::getRayCount();
			Time t = clock.getFrametime();
			cout << "frame: " << clock.getFrameCount()
				<< " time: " << t
				<< " (" << scene->getObjects().size() << " objects, "
				<<  r / (t * 1000.f) << " MRays/s)"<<'\n';

			if (makeScreenshot)
			{
				char screenshotFilename[50];
				sprintf(screenshotFilename, "%d.bmp", clock.getFrameCount());
				screen->save(screenshotFilename);
			}
		}

		isRunning = platform::handleInput(*this);
		screen->update();
		platform::sleep(1); // we don't need all the cpu time :)
	}
	shutdown(*screen.get());
}


bool Engine::parseCommandLine(int argc, char** argv)
{
	CmdLineParser p;

	if (!p.parse(argc, argv))
		return false;

	int numThreads = 1;

	p.getArgument("threads", &numThreads);
	p.getArgument("width", &screenWidth);
	p.getArgument("height", &screenHeight);
	p.getArgument("anti-aliasing", &supersampling);
	p.getArgument("recursiondepth", &recursionDepth);

	inputfile = p.getArgument("data-file");

	if (p.isOptionPresent("video"))
		videofile = p.getArgument("video");

	makeScreenshot = p.isOptionPresent("screenshot");
	showFps = p.isOptionPresent("showfps");
	renderOnlyOneFrame = p.isOptionPresent("oneframe");
#ifdef _OPENMP
	omp_set_num_threads(numThreads);
#endif

	return true;

}


void Engine::saveScene(const std::string& path) const
{
	cout << "saving the current scene to " << path << '\n';
	scene->writeToFile(path);
}


void Engine::tearDown()
{
}
