/**
 * KdTree.cpp
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "KdTree.h"
#include "KdTreeHelper.h"
#include "Scene.h"
#include "AxisAlignedBoundingBox.h"
#include "util/Vector3D.h"
#include "Ray.h"
#include "IntersectionData.h"
#include "defaultsettings.h"

#include <cassert>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

using util::Vector3D;
using namespace blurt;

const size_t MAX_STACK_SIZE = 50; // size of the stack used for the traversal


blurt::BaseKdTree::BaseKdTree(Scene& s)
	: rootNode(0), scene(s)
{
}


blurt::BaseKdTree::~BaseKdTree()
{
	if (rootNode)
		deleteTree(rootNode);
}

void blurt::BaseKdTree::deleteTree(KdNode* node)
{
	delete node; // the KdNode-Dtor will take care of the rest
}


// implemented after TAb_rec in Havran's PhD
void blurt::BaseKdTree::traceRay(const Ray& ray, IntersectionData* id) const
{
	assert(rootNode);
	float near = 0, far = 0;

	if (!rootNode->intersectRay(scene.getBounds(), ray, &near, &far))
		return;

	const Vector3D& origin = ray.getOrigin();
	const Vector3D& direction = ray.getDirection();

	StackElement stack[MAX_STACK_SIZE];

	KdNode* currentNode = rootNode;

	size_t entrypointIdx = 0;
	stack[entrypointIdx].t = near;

	if (near > 0.0f) // external entrypoint
		stack[entrypointIdx].pb = origin + near * direction;
	else
		stack[entrypointIdx].pb = origin;

	size_t exitpointIdx = 1;
	stack[exitpointIdx].t = far;
	stack[exitpointIdx].pb = origin + far * direction;
	stack[exitpointIdx].node = 0;

	KdNode* farNode;
	while (currentNode != 0)
	{
		while (!currentNode->isLeaf())
		{
			float splitpos = currentNode->getSplitPosition();
			int axis = currentNode->getSplitAxis();

			if (stack[entrypointIdx].pb[axis] <= splitpos)
			{
				if (stack[exitpointIdx].pb[axis] <= splitpos)
				{
					currentNode = currentNode->getLeft();
					continue;
				}
				if (stack[exitpointIdx].pb[axis] == splitpos)
				{
					currentNode = currentNode->getRight();
					continue;
				}
				farNode = currentNode->getRight();
				currentNode = currentNode->getLeft();
			}
			else
			{
				if (stack[exitpointIdx].pb[axis] > splitpos)
				{
					currentNode = currentNode->getRight();
					continue;
				}
				farNode = currentNode->getLeft();
				currentNode = currentNode->getRight();
			}

			float t = (splitpos - origin[axis]) / direction[axis];
			int tmp = exitpointIdx++;
			if (exitpointIdx == entrypointIdx) ++exitpointIdx;
			assert(exitpointIdx < MAX_STACK_SIZE && "stack to small in kdtree traversal");

			// push new values on the stack
			static const int nextAxis[] = {1, 2, 0};
			static const int prevAxis[] = {2, 0, 1};
			stack[exitpointIdx].prev = tmp;
			stack[exitpointIdx].t = t;
			stack[exitpointIdx].node = farNode;
			stack[exitpointIdx].pb[axis] = splitpos;
			int nextaxis = nextAxis[axis];
			int prevaxis = prevAxis[axis];
			stack[exitpointIdx].pb[nextaxis] = origin[nextaxis] + t * direction[nextaxis];
			stack[exitpointIdx].pb[prevaxis] = origin[prevaxis] + t * direction[prevaxis];
		}

		const kdtree::ObjectList& list = currentNode->getObjectList();

		float distance = 0.0f;
		bool isIntersection = false;

		foreach (const TraceableObject* obj, list)
		{
			isIntersection = obj->intersectsRay(ray, &distance);
			if (distance >= stack[entrypointIdx].t && distance <= stack[exitpointIdx].t)
				id->testAndSetHit(obj, distance);
		}

		if (id->isIntersection())
			return;

		entrypointIdx = exitpointIdx;
		currentNode = stack[exitpointIdx].node;
		exitpointIdx = stack[entrypointIdx].prev;
	}
}


bool blurt::BaseKdTree::isFirstIntersection(const Ray& ray, float distance) const
{
	IntersectionData id;
	traceRay(ray, &id);
	return id.getDistance() >= distance;
}
