#ifndef KDTREE_H
#define KDTREE_H
/**
 * KdTree.h
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "util/Vector3D.h"
using util::Vector3D;

namespace blurt
{

class TraceableObject;
class Scene;
class IntersectionData;
class Ray;
class AxisAlignedBoundingBox;


namespace kdtree
{
	class KdNode;
};

/*
 * Baseclass for the KdTree.
 * The "real" KdTree class can be parametritized with a SplitPositionCalculator.
 * Everthing else is contained in this class so we don't have to inline all the
 * functions.
 */
class BaseKdTree
{
	public:
		BaseKdTree(Scene& s);
		~BaseKdTree();

		void build();
		void optimizeScene();
		void traceRay(const Ray& ray, IntersectionData* id) const;
		bool isFirstIntersection(const Ray& ray, float distance) const;

	protected:
		typedef kdtree::KdNode KdNode;

		void buildRecursivly(const AxisAlignedBoundingBox& voxel,
		                     KdNode* currentNode, size_t level);

		void deleteTree(KdNode* node);

		/// Stack used for the traversal
		struct StackElement
		{
			KdNode* node;  // pointer to the far child
			float t;       // signed entry/exit distance
			Vector3D pb;   // the coordinates of entry/exit point
			int prev;      // the pointer to the previous stack item
		};

		KdNode* rootNode;
		Scene& scene;
};



template <class SplitPlanePositioner>
class KdTree : public BaseKdTree
{
	public:
		KdTree(Scene& s) : BaseKdTree(s) {}
		void build();
		void optimizeScene();
	private:
		void buildRecursivly(const AxisAlignedBoundingBox& voxel,
		                     KdNode* currentNode, size_t level);

};

};

#include "KdTreeHelper.h"
#include "Scene.h"


template <class T>
inline void blurt::KdTree<T>::build()
{
	optimizeScene(); // rebuild the scene
}


template <class T>
inline void blurt::KdTree<T>::optimizeScene()
{
	if (rootNode)
		deleteTree(rootNode);

	rootNode = new KdNode();
	foreach(const TraceableObject& o, scene.getObjects())
		rootNode->addObject(o);

	buildRecursivly(scene.getBounds(), rootNode, 0);
}


const size_t MAX_TREE_DEPTH = 20;
const size_t MAX_OBJECTS_IN_LEAF = 4;

template <class SplitPlanePositioner>
inline void blurt::KdTree<SplitPlanePositioner>::buildRecursivly(const AxisAlignedBoundingBox& voxel, KdNode* currentNode, size_t level)
{
	if (level >= MAX_TREE_DEPTH)
	{
//		std::cout << "max depth reached\n";
		return;
	}
	++level;

	// determine axis;
	Vector3D size(voxel.pos2 - voxel.pos1);
	unsigned char axis = (size[1] > size[0]) ? 1 : 0;;
	axis = (size[2] > size[axis]) ? 2 : axis;

	currentNode->setSplitAxis(axis);


	// find the splitposition
	float splitpos = SplitPlanePositioner::getSplitPosition(voxel, size, axis, currentNode);
	currentNode->setSplitPosition(splitpos);
//std::cout << "level: " << level << " axis: " << (int) axis << " pos: " << splitpos << '\n';

	// calculate the new voxels
	AxisAlignedBoundingBox newBounds[2];
	newBounds[0].pos1 = voxel.pos1;
	newBounds[0].pos2 = voxel.pos2;
	newBounds[1].pos1 = voxel.pos1;
	newBounds[1].pos2 = voxel.pos2;
	newBounds[0].pos2[axis] = currentNode->getSplitPosition();// + EPSILON;
	newBounds[1].pos1[axis] = newBounds[0].pos2[axis];// - EPSILON;

	currentNode->setLeft(new KdNode());
	currentNode->setRight(new KdNode());

	size_t nLeft = 0, nRight = 0;
	const kdtree::ObjectList& l = currentNode->getObjectList();

	foreach (const TraceableObject* obj, l)
	{
		if (obj->intersectsBox(newBounds[0].pos1, newBounds[0].pos2))
		{
			currentNode->getLeft()->addObject(*obj);
			++nLeft;
		}

		if (obj->intersectsBox(newBounds[1].pos1, newBounds[1].pos2))
		{
			currentNode->getRight()->addObject(*obj);
			++nRight;
		}
	}

	if (nRight + nLeft > MAX_OBJECTS_IN_LEAF)
	{
//std::cout << "level: " << level << " left: " << nLeft << " right: " << nRight << '\n';
		buildRecursivly(newBounds[0], currentNode->getLeft(), level);
		buildRecursivly(newBounds[1], currentNode->getRight(), level);
	}
	else
	{
//std::cout << "max objects criterion reached in level " << level << '\n';
		delete currentNode->getLeft();
		currentNode->setLeft(0);
		delete currentNode->getRight();
		currentNode->setRight(0);
	}
}


#endif // KDTREE_H
