/**
 * KdTreeHelper.cpp
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "KdTreeHelper.h"
#include "AxisAlignedBoundingBox.h"
#include "Ray.h"
#include "TraceableObject.h"
#include "defaultsettings.h"

using namespace blurt;


float blurt::kdtree::SpatialMedianSplitter::getSplitPosition(const AxisAlignedBoundingBox& voxel,
							                  const Vector3D& size, Axis a,
							                  KdNode* currentNode)
{
	return voxel.pos1[a] + size[a] / 2;
}



float blurt::kdtree::SahSplitter::getSplitPosition(const AxisAlignedBoundingBox& voxel,
	                                        const Vector3D& size, Axis a,
	                                        KdNode* currentNode)
{
	float bestpos = -1;
	float bestcost = std::numeric_limits<float>::max();

	// calculate surface area for current node
	float sav = size[0] * size[1] + size[0] * size[2] + size[1] * size[2];
	float invSav = 1.0f / sav;

	const ObjectList& list = currentNode->getObjectList();

	float cost;
	foreach (const TraceableObject* o, list)
	{
		float left = o->getAxisAlignedBoundingBox().pos1[a] - EPSILON;
		cost = calculateCost(left, list, a, invSav);
		if (cost < bestcost)
		{
			bestcost = cost;
			bestpos = left;
		}

		float right = o->getAxisAlignedBoundingBox().pos2[a] + EPSILON;
		cost = calculateCost(right, list, a, invSav);
		if (cost < bestcost)
		{
			bestcost = cost;
			bestpos = right;
		}
	}

	return bestpos;
}


/*
 * Calculates the cost for the given splitpos.
 */
float blurt::kdtree::SahSplitter::calculateCost(float splitpos, ObjectList l, Axis a, float invSav)
{

	float leftArea = 0.f;
	float rightArea = 0.f;
	unsigned int leftCount = 0;
	unsigned int rightCount = 0;


	foreach (const TraceableObject* o, l)
	{
		const AxisAlignedBoundingBox& aabb = o->getAxisAlignedBoundingBox();
		const Vector3D size(aabb.pos2 - aabb.pos1);
		float sa = size[0] * size[1]  + size[0] * size[2] + size[1]*size[2];
		if (aabb.pos1[a] < splitpos || aabb.pos2[a] < splitpos)
		{
			leftArea += sa;
			++leftCount;
		}

		if (aabb.pos2[a] > splitpos || aabb.pos1[a] > splitpos)
		{
			rightArea += sa;
			++rightCount;
		}
	}

	float tmp = leftArea * leftCount + rightArea * rightCount;

	// 0.3 and 1.0 are just chosen for the traversal and intersection
	// after Phantom's raytracing-tutorial
	return 0.3f + 1.0f * invSav * tmp;
}



blurt::kdtree::KdNode::~KdNode()
{
	delete left;
	delete right;
}


bool blurt::kdtree::KdNode::intersectRay(const AxisAlignedBoundingBox& bounds,
                                    const Ray& ray, float* near, float* far)
{
	*near = -std::numeric_limits<float>::max();
	*far = std::numeric_limits<float>::max();

	const Vector3D& dir = ray.getDirection(); // let's save us some typing ;p
	const Vector3D& origin = ray.getOrigin();

	for( int i = 0; i < 3; ++i )
	{
		float tmpNear, tmpFar;

		if (dir[i] > 0.0f)
		{
			const float invDir = 1.0f / dir[i];
			tmpNear = (bounds.pos1[i] - origin[i]) * invDir;
			tmpFar = (bounds.pos2[i] - origin[i]) * invDir;
		}
		else if (dir[i] < 0.0f)
		{
			const float invDir = 1.0f / dir[i];
			tmpNear = (bounds.pos2[i] - origin[i] ) * invDir;
			tmpFar = (bounds.pos1[i] - origin[i] ) * invDir;
		}
		else
		{
			if (origin[i] < bounds.pos1[i] || origin[i] > bounds.pos2[i])
				return false;
			else
				continue;
		}

		if (*near > tmpFar || *far < tmpNear)
			return false;

		if (tmpNear > *near) *near = tmpNear;
		if (tmpFar < *far) *far = tmpFar;
	}
	return true;
}


