/**
 * ParticleSystem.cpp: a simple ParticleSystem.
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "ParticleSystem.h"
#include "Sphere.h"
#include "defaultsettings.h"
#include "Material.h"
#include "Engine.h"
#include "Scene.h"
#include <sstream>


using namespace blurt;

const Vector3D DEFAULT_ACCELERATION(0, -1.5, 0);
const Vector3D DEFAULT_POS_VARIANCE(1, 1, 1);
const Vector3D DEFAULT_VEL_VARIANCE(7.7, 1.7, 1.7);
const float PARTICLE_SIZE = 0.2;
const float DEFAULT_TTL = 1.52;
const float DEFAULT_SPAWNRATE = 2500;
const float DELTA_T = 1.f / LOGICTICKS_PER_SECOND;

namespace
{
	// returns a random Number between -1 and 1
	const float RND_MAX = (float) RAND_MAX;
	inline float random()
	{
		return (2 * rand() / RND_MAX) - 1.0f;
	}

	// returns a vector that lies somwhere in a distance of max "var" from "pos"
	inline Vector3D randomVector(const Vector3D& pos, const Vector3D& var)
	{
		Vector3D retval(pos);
		retval[0] += random() * var[0];
		retval[1] += random() * var[1];
		retval[2] += random() * var[2];
		return retval;
	}
};


blurt::ParticleSystem::ParticleSystem(const Vector3D& pos, const Vector3D& emitDirection)
	: position(pos),
	  positionVariance(DEFAULT_POS_VARIANCE),
	  emittingDirection(emitDirection),
	  velocityVariance(DEFAULT_VEL_VARIANCE),
	  acceleration(DEFAULT_ACCELERATION),
	  ttl(DEFAULT_TTL),
	  spawnrate(DEFAULT_SPAWNRATE)
{
	const char* materialDesc = "MATERIAL(Color: (0.9, 0, 0)	Diffuse: 0.99	Specular: 0.01	Phong: 1	Reflective: 0.2	Refractive: 0.8	RefractionIndex: 1.3	Density: 0.15)";
	std::istringstream tmp(materialDesc);
	tmp >> material;
}


blurt::ParticleSystem::ParticleSystem(std::istream& is)
{
	// the first line is empty, so discard it
//	std::string tmp;
//	getline(is, tmp);

	is >> material;
	is.ignore(11);
	is >> position;
	is.ignore(11);
	is >> positionVariance;
	is.ignore(12);
	is >> emittingDirection;
	is.ignore(11);
	is >> velocityVariance;
	is.ignore(15);
	is >> acceleration;
	is.ignore(5);
	is >> ttl;
	is.ignore(11);
	is >> spawnrate;
}


std::ostream& blurt::operator<< (std::ostream& os, const ParticleSystem& ps)
{
	os << ps.material << '\n';
	os << "Position: " << ps.position
	   << "\tVariance: " << ps.positionVariance << '\n';
	os << "Direction: " << ps.emittingDirection
	   << "\tVariance: " << ps.velocityVariance << '\n';
	os << "Acceleration: " << ps.acceleration << '\n';
	os << "TTL: " << ps.ttl << "\tSpawnrate: " << ps.spawnrate;
	return os;
}


blurt::ParticleSystem::~ParticleSystem()
{
}



void blurt::ParticleSystem::update()
{
	createParticles();

	for (ParticleIter i = particles.begin(); i != particles.end(); ++i)
	{
		i->pos += i->vel * DELTA_T;
		i->vel += acceleration * DELTA_T;

		i->sphere->setCenter(i->pos);
		--i->ttl;

		// remove dead particles
		if (i->ttl <= 0)
		{
			Scene& s = Engine::get().getScene();
			s.removeObject(i->sphere);
			i = particles.erase(i);
		}
	}

}


void blurt::ParticleSystem::createParticles()
{
	const unsigned spawnsPerFrame = spawnrate * DELTA_T;

	Scene& s = Engine::get().getScene();
	for (size_t i = 0; i < spawnsPerFrame; ++i)
	{
		Particle p;
		p.pos = randomVector(position, positionVariance);
		p.vel = randomVector(emittingDirection, velocityVariance);
		p.sphere = new Sphere("particle", material, p.pos, PARTICLE_SIZE);
		p.ttl = ttl * LOGICTICKS_PER_SECOND;
		s.addObject(p.sphere);
		particles.push_back(p);
	}
}



