/**
 * Ray.cpp
 *
 * (c) 2007, 2008 Thomas Unterthiner
 *
 */

#include "RayPacket.h"
#include <algorithm>
#include "blurt/Ray.h"
#include "blurt/IntersectionData.h"
#include "util/Color.h"

#include "TraceableObject.h"

using util::sse::VectorArray3D;

namespace blurt
{


RayPacket::RayPacket(const Vector3D* origins, const Vector3D* dirs, size_t numRays)
	: directions(dirs, numRays),
	  origins(origins, numRays),
	  distances(std::numeric_limits<float>::max())
{
	assert(numRays == PACKET_SIZE && "only up to 4 rays per packet allowed!");
	directions.normalize();
	std::fill_n(hitObjects.begin(), numRays, (TraceableObject*) 0);
}


void RayPacket::resetIntersectionData()
{
	distances = VectorSSE(std::numeric_limits<float>::max());
	std::fill_n(hitObjects.begin(), PACKET_SIZE, (TraceableObject*) 0);
}


Ray RayPacket::getRay(size_t idx)
{
	assert(idx < PACKET_SIZE && "only up to 4 rays per packet allowed!");

	return Ray(
		Vector3D(origins[0][idx], origins[1][idx], origins[2][idx]),
		Vector3D(directions[0][idx], directions[1][idx], directions[2][idx]));
}


RayArray RayPacket::getRays() const
{
	const VectorArray3D o = convertToVector3D(origins);
	const VectorArray3D d = convertToVector3D(directions);

	RayArray retval = {{Ray(o[0], d[0]), Ray(o[1], d[1]),
	                    Ray(o[2], d[2]), Ray(o[3], d[3])}};
	return retval;
}


IntersectionData RayPacket::getIntersectionData(size_t idx) const
{
	IntersectionData id;
	id.setHit(hitObjects[idx], distances[idx]);
	return id;
}

void RayPacket::setIntersectionData(size_t idx, float distance,
                                    const TraceableObject* hitObject)
{
	distances[idx] = distance;
	hitObjects[idx] = hitObject;
}


void RayPacket::updateHitInformation(const VectorSSE& newDists,
                                     const VectorSSE& validDistsMask,
                                     const TraceableObject* obj)
{
	const VectorSSE& resultMask = compareLt(newDists, distances) & validDistsMask;

	for (unsigned i = 0; i < RayPacket::PACKET_SIZE; ++i)
	{
		if (!resultMask.isZero(i))
		{
			setHitObject(i,obj);
			distances[i] = newDists[i];
		}
	}
}

}



