/**
 * Raytracer.cpp
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Raytracer.h"
#include "Ray.h"
#include "Scene.h"
#include "Sphere.h"
#include "Engine.h"
#include "TraceableObject.h"
#include "RayPacket.h"
#include "../defaultsettings.h"
#include "IntersectionData.h"
#include "IntersectionInformation.h"
#include "platform/Framework.h"
#include "util/TheoraEncoder.h"
#include "util/mathhelper.h"
#include "util/Color.h"
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH

#ifdef _OPENMP
#include <omp.h>
#endif

using namespace blurt;
using namespace util;
using namespace platform;

// how many rows should one thread process
#define CHUNKS 32

const float INITIAL_REFRACTION_INDEX = 1.f;
bool DEBUGINFO = false;

namespace blurt
{


Raytracer::Raytracer(platform::Screen& screen_, size_t maxRecursionDepth_)
	: screen(screen_),
	  scene(Engine::get().getScene()),
	  spatialSubdivision(Engine::get().getScene()),
	  maxRecursionDepth(maxRecursionDepth_),
	  screenWidth(screen.getWidth()),
	  screenHeight(screen.getHeight()),
	  dx(scene.getViewportWidth()/screenWidth),
	  dy(scene.getViewportHeight()/screenHeight),
	  precalculatedRays(0),
	  precalculatedDirections(0),
	  precalculatedRayPackets(0),
	  image(0)
{
	precalculateRays();
	precalculateRayPackets();
	image = new Color[screenWidth * screenHeight];
	spatialSubdivision.optimizeScene();
}


Raytracer::~Raytracer()
{
	// add the last videoframe with isLastFrame==true
	if (video.get())
		video->addFrame(screenWidth, screenHeight, image, true);

	if (precalculatedRays)
		delete[] precalculatedRays;
	if (image)
		delete[] image;
}



void Raytracer::enableRenderToVideo(const char* videofile)
{
	video.reset(new TheoraEncoder(videofile, screenWidth, screenHeight, LOGICTICKS_PER_SECOND));
}


Color Raytracer::traceRay(const Ray& ray, float* distance,
                          size_t recursionDepth, float currentRefraction)
{
	IntersectionData id;
	spatialSubdivision.traceRay(ray, &id);
	*distance = id.getDistance();

	return calculateColor(ray, id, recursionDepth, currentRefraction);
}


void Raytracer::traceRayPacket(RayPacket* rp, size_t recursionDepth,
		                       float currentRefraction, Color returnValues[])
{
	spatialSubdivision.traceRayPacket(rp);
	RayArray rays = rp->getRays();
	for (unsigned i = 0; i < RayPacket::PACKET_SIZE; ++i)
	{
		IntersectionData id = rp->getIntersectionData(i);
//		if (DEBUGINFO) std::cout << (id.getHitObject() == 0) << '\n';
		returnValues[i] = calculateColor(rays[i], id, recursionDepth, currentRefraction);
	}
}


Color Raytracer::calculateShading(const Ray& ray, const IntersectionInformation& ii)
{
	Color c(0, 0, 0);
	const Vector3D& normal = ii.getIntersectionNormal();
	const Vector3D& intersectionPoint = ii.getIntersectionPoint();
	const TraceableObject& obj = *ii.getHitObject();
	const Material& m = obj.getMaterial();


	foreach (const Light& light, scene.getLights())
	{
		Vector3D l = light.getCenter() - intersectionPoint;
		float distToLight = l.getLength();
		l.normalize();
		float dotLightNormal = dotProduct(l, normal);

		// shadowing
		float shadow = 1.0f;
		if (light.getType() == TraceableObject::SPHERE)
		{
			Ray r(intersectionPoint + EPSILON * l, l);

			if (!spatialSubdivision.isFirstIntersection(r, distToLight))
				shadow = 0.0f;
		}

		if (shadow < 0.01f)
			continue;

		// diffuse shading
		if (m.getDiffuse() > 0)
		{
			if (dotLightNormal > 0)
			{
				float diff = dotLightNormal * m.getDiffuse() * shadow;
				c += diff * m.getColor() * light.getMaterial().getColor();
			}
		}

		// specular shading
		if (m.getSpecular() > 0)
		{
			const Vector3D& v = ray.getDirection();
			Vector3D r = l - 2.0f * dotLightNormal * normal;
			float dot = dotProduct(v, r);
			if (dot > 0.0f)
			{
				float pc = m.getPhongConstant();
				float spec = pow(dot, pc) * m.getSpecular() * shadow;
				c += spec * m.getColor();
			}
		}
	}
	return c;
}


Color Raytracer::calculateReflection(const Ray& ray, const IntersectionInformation& ii,
                                     size_t recursionDepth,
                                     float currentRefraction)
{
	const Vector3D& normal = ii.getIntersectionNormal();
	const Vector3D& intersectionPoint = ii.getIntersectionPoint();
	const TraceableObject& obj = *ii.getHitObject();
	const Material& m = obj.getMaterial();

	Color c(0, 0, 0);
	float refl = m.getReflective();
	if (refl > 0.0f)
	{
		Vector3D r = ray.getDirection() -
			2.0f * dotProduct(ray.getDirection(), normal) * normal;
		if (recursionDepth < maxRecursionDepth)
		{
			float tmp;
			Ray start(intersectionPoint + EPSILON*r, r);
			Color reflCol(traceRay(start, &tmp,
				recursionDepth + 1, currentRefraction));
			c += refl * reflCol; // * m.getColor();
		}
	}
	return c;
}


Color Raytracer::calculateRefraction(const Ray& ray,
									 const IntersectionInformation& ii,
							         float currentRefraction,
							         size_t recursionDepth)
{
	const Vector3D& normal = ii.getIntersectionNormal();
	const Vector3D& intersectionPoint = ii.getIntersectionPoint();
	const TraceableObject& obj = *ii.getHitObject();
	const Material& m = obj.getMaterial();

	Color c(0, 0, 0);
	float refr = m.getRefractive();
	if (refr > 0 && recursionDepth < maxRecursionDepth)
	{
		float ri =  m.getRefractionIndex();
		float n = currentRefraction / ri;

		// TODO: not sure if we need the following line now that we
		// don't separate bewteen OUTSIDE and INSIDE intersections
		Vector3D norm = (ii.isIntersection() ? 1.f : -1.f) * normal;
		float cosI = -dotProduct(norm, ray.getDirection());
		float cosT2 = 1.0f - n * n * (1.0f - cosI * cosI);
		if (cosT2 > 0.0f)
		{
			Vector3D t = (n * ray.getDirection()) + (n * cosI - sqrtf( cosT2 )) * norm;
			float distance;
			Ray r(intersectionPoint + EPSILON*t, t);
			Color refrCol = traceRay(r, &distance, recursionDepth + 1, ri);

			// Beer's Laws
			Color absorbance = -distance * m.getDensity() * m.getColor();
			Color transparency = Color( exp( absorbance[0] ),
										exp( absorbance[1] ),
										exp( absorbance[2] ) );
			c += refrCol * transparency;
//			c+= refrCol;
		}
	}
	return c;
}


void Raytracer::render()
{
#if 0
	renderPackets();
#else
	const Vector3D& cameraPos = scene.getCameraPosition();

	if (cameraPos != lastCameraPosition)
		precalculateRays();
	spatialSubdivision.build();
	float tmp;

	// we traverse in subimages of 32x32
	// (optimizes chache-coherence a bit)
	const int pixelD = 1;

	#pragma omp parallel for default(shared) private (tmp) schedule(static, CHUNKS)
	for (int y = 0; y < screenHeight; y += pixelD)
	{
		for (int x = 0; x < screenWidth; x += pixelD)
		{

			for (int yi = y; yi < y + pixelD; ++yi)
			{
				for (int xi = x; xi < x + pixelD; ++xi)
				{
					unsigned int pos = yi * screenWidth + xi;
					Color c = traceRay(precalculatedRays[pos],
					                   &tmp, 0, INITIAL_REFRACTION_INDEX);
					image[pos] = c;
				}
			}
		}
	}
	screen.draw(image);
	if (video.get())
		video->addFrame(screenWidth, screenHeight, image, false);
#endif
}


void Raytracer::renderPackets()
{
	const Vector3D& cameraPos = scene.getCameraPosition();

	if (cameraPos != lastCameraPosition)
			precalculateRayPackets();
	spatialSubdivision.build();

	// packetsizeSquareRoot = squareroot of Ray::PACKET_SIZE
	const int packetsizeSquareRoot = 2;
	const int PACKET_SIZE = RayPacket::PACKET_SIZE;

//	#pragma omp parallel for default(shared) schedule(static, CHUNKS)
	for (int y = 0; y < screenHeight; y += packetsizeSquareRoot)
	{
		for (int x = screenWidth - 1; x >= 0; x -= packetsizeSquareRoot)
		{
			const unsigned pos = (y * screenWidth) / 4 + (x/2);

			Color colorReturn[PACKET_SIZE];
			RayPacket* rp = &precalculatedRayPackets[pos];
			rp->resetIntersectionData();
//			traceRayPacket(rp,0, INITIAL_REFRACTION_INDEX, colorReturn);

			// a single RayPacket stores the ray for (x,y), (x-1,y),
			// (x, y+1), (x-1, y+1)
			const unsigned imagePos = y * screenWidth + x;
			image[imagePos] = colorReturn[0];
			image[imagePos-1] = colorReturn[1];
			image[imagePos+screenWidth] = colorReturn[2];
			image[imagePos+screenWidth-1] = colorReturn[3];
		}
	}
	screen.draw(image);
	if (video.get())
		video->addFrame(screenWidth, screenHeight, image, false);
}


void Raytracer::renderSupersampled(int samples)
{
	float posX = scene.getViewportX();
	float posY = scene.getViewportY();

	float sdx = dx / samples;
	float sdy = dy / samples;

	const Vector3D& cameraPos = scene.getCameraPosition();
	spatialSubdivision.build();

	int i, j;
	float tmp;

	#pragma omp parallel for default(shared) \
		private(i, j, tmp) private(posX, posY) schedule(static, CHUNKS)
	for (int y = 0; y < screenHeight; ++y)
	{
		// to make the parallelisation work we can't use the usual
		// pos += delta increments, but have to initialize the position
		// variables properly
		posX = scene.getViewportX();
		posY = scene.getViewportY() + y * dy;

		for (int x = 0; x < screenWidth; ++x)
		{
			Color c(0, 0, 0);
			for (i = 0; i < samples; ++i)
			{
				for (j = 0; j < samples; ++j)
				{
					Vector3D dir(posX + i*sdx, posY + j*sdy, 0);
					dir -= cameraPos;
					dir.normalize();
					Ray ray(cameraPos, dir);

					c += traceRay(ray, &tmp, 0, INITIAL_REFRACTION_INDEX);
				}
			}
			c /= samples*samples;
			image[y * screenWidth + x] = c;
			posX += dx;
		}
	}
	screen.draw(image);
	if (video.get())
		video->addFrame(screenWidth, screenHeight, image, false);
}


void Raytracer::precalculateRays()
{
	if (precalculatedRays)
		delete[] precalculatedRays;

	const Vector3D& cameraPos = scene.getCameraPosition();
	float posX = scene.getViewportX();
	float posY = scene.getViewportY();

	precalculatedRays = new Ray[screenHeight * screenWidth];
	#pragma omp parallel for schedule(static, CHUNKS) ordered
	for (int y = 0; y < screenHeight; ++y)
	{
		for (int x = screenWidth - 1; x >= 0; --x)
		{
			Vector3D dir(posX, posY, 0);
			dir -= cameraPos;
			dir.normalize();
			Ray ray(cameraPos, dir);
			precalculatedRays[y * screenWidth + x] = ray;
			#pragma omp ordered
			{
				posX += dx;
			}
		}
		#pragma omp ordered
		{
			posY += dy;
			posX = scene.getViewportX();
		}
	}
	lastCameraPosition = cameraPos;
}

void Raytracer::precalculateDirections()
{
	if (precalculatedDirections)
		delete[] precalculatedDirections;

	const Vector3D& cameraPos = scene.getCameraPosition();
	float posX = scene.getViewportX();
	float posY = scene.getViewportY();

	precalculatedDirections = new Vector3D[screenHeight * screenWidth];
	#pragma omp parallel for schedule(static, CHUNKS) ordered
	for (int y = 0; y < screenHeight; ++y)
	{
		for (int x = screenWidth - 1; x >= 0; --x)
		{
			Vector3D dir(posX - cameraPos[0], posY - cameraPos[1], -cameraPos[2]);
			precalculatedDirections[y * screenWidth + x] = dir;
			#pragma omp ordered
			{
				posX += dx;
			}
		}
		#pragma omp ordered
		{
			posY += dy;
			posX = scene.getViewportX();
		}
	}
	lastCameraPosition = cameraPos;
}


void Raytracer::precalculateRayPackets()
{
	if (precalculatedRayPackets)
		delete[] precalculatedRayPackets;
	precalculateDirections();
	const Vector3D& cameraPos = scene.getCameraPosition();


	// the following code (and addressing in the precalculatedRayPacket
	// array in general assume that there are 4 Packets per ray!
	assert(RayPacket::PACKET_SIZE == 4 && "expecting 4 rays per packet");
	const size_t size = (screenHeight * screenWidth) / 4;
	precalculatedRayPackets = new RayPacket[size];

	// packetsizeSquareRoot = squareroot of Ray::PACKET_SIZE
	const int packetsizeSquareRoot = 2;
	const int PACKET_SIZE = RayPacket::PACKET_SIZE;
	#pragma omp parallel for schedule(static, CHUNKS) ordered
	for (int y = 0; y < screenHeight; y += packetsizeSquareRoot)
	{
		for (int x = screenWidth - 1; x >= 0; x -= packetsizeSquareRoot)
		{
			Vector3D dirs[PACKET_SIZE];
			Vector3D origins[PACKET_SIZE];
			int i = 0;
			for (int yi = y; yi < y + packetsizeSquareRoot; ++yi)
			{
				for (int xi = x; xi > x - packetsizeSquareRoot; --xi)
				{
					const unsigned pos = yi * screenWidth + xi;
					dirs[i] = precalculatedDirections[pos];
					origins[i] = cameraPos;
					++i;
				}
			}
			const unsigned pos = (y * screenWidth) / 4 + x/2;
			precalculatedRayPackets[pos] = RayPacket(origins, dirs, PACKET_SIZE);
		}
	}
	lastCameraPosition = cameraPos;
}
}

