#ifndef RAYTRACER_H
#define RAYTRACER_H

/**
 * Raytracer.h: the actual raytracer
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "util/Vector3D.h"
#include "util/Color.h"
#include "IntersectionData.h"
#include "IntersectionInformation.h"
#include "SpatialSubdivision.h"
#include <boost/scoped_ptr.hpp>


namespace platform {class Screen;}
namespace util {class TheoraEncoder;}

namespace blurt
{

using util::Vector3D;
using util::Color;


class Ray;
class RayPacket;
class TraceableObject;
class Scene;
class Material;
class IntersectionData;
class IntersectionInformation;

class Raytracer
{
	public:
		Raytracer(platform::Screen& screen, size_t maxRecursionDepth);
		~Raytracer();

		void render();

		void renderPackets();

		void enableRenderToVideo(const char* videofile);

		/**
		 * renders with samples*samples supersamples per pixel :)
		 */
		void renderSupersampled(int samples);

	private:

		/**
		 * returns the color that comes from this ray,
		 * and the distance of the ray has travelled
		 */
		Color traceRay(const Ray&, float* distance,
			size_t recursionDepth, float currentRefraction);

		void traceRayPacket(RayPacket* rp,
		                    size_t recursionDepth,
		                    float currentRefraction,
		                    Color returnValues[]);

		Color calculateColor(const Ray& ray, const IntersectionData& id,
		                     size_t recursionDepth, float currentRefraction);

		Color calculateShading(const Ray& ray, const IntersectionInformation& id);

		Color calculateReflection(const Ray& ray, const IntersectionInformation& id,
                                  size_t recursionDepth,
                                  float currentRefraction);

		Color calculateRefraction(const Ray& ray,
							      const IntersectionInformation& id,
							      float currentRefraction,
							      size_t recursionDepth);

		void precalculateRays();
		void precalculateRayPackets();
		void precalculateDirections();

		platform::Screen& screen;
		const Scene& scene;
		SpatialSubdivision spatialSubdivision;

		size_t maxRecursionDepth;

		// viewport size in screen coordinates
		int screenWidth;
		int screenHeight;

		// distance from one ray to the next on the viewport
		float dx;
		float dy;

		// in case the camera doesn't change we can precalculate
		// the Rays.
		Ray* precalculatedRays;
		Vector3D* precalculatedDirections;
		RayPacket* precalculatedRayPackets;
		Color* image;
		Vector3D lastCameraPosition;

		boost::scoped_ptr<util::TheoraEncoder> video;
};


inline Color Raytracer::calculateColor(const Ray& ray, const IntersectionData& id,
                                size_t recursionDepth, float currentRefraction)
{
	const Color BACKGROUND_COLOR(1, 1, 1);
	if (!id.getHitObject())
		return BACKGROUND_COLOR;

	IntersectionInformation ii(id, ray);
	Color shade = calculateShading(ray, ii);
	Color reflection = calculateReflection(ray, ii, recursionDepth, currentRefraction);
//	Color refraction = calculateRefraction(ray, ii, currentRefraction, recursionDepth);

	return shade + reflection;// + refraction;
}

};

#endif // RAYTRACER_H
