#include "RegularGrid.h"
#include "Engine.h"
#include "Scene.h"
#include "Ray.h"
#include "IntersectionData.h"
#include "defaultsettings.h"
#include "Triangle.h"
#include <cstring>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH


using namespace blurt;
using namespace util;

/*
 * NOTE:
 * For some strange reason there are artefacts in the picture (spheres left out or cut to pieces)
 * if the total gridsize isn't a multiple of the cellsize.
 * Usually the gridsize is defined by what the plane's return as their AABB, so we're
 * on the safe side for now. But this issue definitely would need some investigating ;)
 */
const float GRID_CELLSIZE = 2;


blurt::RegularGrid::RegularGrid(Scene& s)
	: grid(0),
	  scene(s),
	  pos(scene.getBounds().pos1),
	  size(scene.getBounds().getLongestEdge(),
	       scene.getBounds().getLongestEdge(),
	       scene.getBounds().getLongestEdge()),
	  p2(scene.getBounds().pos1 + size),
	  cellSize(GRID_CELLSIZE),
	  invCellSize(1.f / cellSize),
	  cellsPerAxis(scene.getBounds().getLongestEdge() * invCellSize + 1),
	  sceneBounds(Engine::get().getScene().getBounds())
{
	const unsigned N = cellsPerAxis * cellsPerAxis * cellsPerAxis;
	grid = new ObjectNode*[N];

	memset(grid, 0, N * sizeof(ObjectNode*));
}


blurt::RegularGrid::~RegularGrid()
{
	deleteGrid();
	delete[] grid;
}

void blurt::RegularGrid::deleteGrid()
{
	const unsigned N = cellsPerAxis * cellsPerAxis * cellsPerAxis;
	ObjectNode* prev = 0;
	for (unsigned i = 0; i < N; ++i)
	{
		ObjectNode* n = grid[i];
		while (n)
		{
			prev = n;
			n = n->next;
			delete prev;
		}
	}

	memset(grid, 0, N * sizeof(ObjectNode*));
}


/*
 * Splits big triangles in smaller pieces
 */
void blurt::RegularGrid::optimizeScene()
{
	// determine a good cellsize
#if 1
	avgObjectLength = 0.0f;
	foreach (const TraceableObject& o, scene.getObjects())
	{
		const AxisAlignedBoundingBox& aabb = o.getAxisAlignedBoundingBox();
		const Vector3D size(aabb.pos2 - aabb.pos1);
		avgObjectLength += (size[0] + size[1] + size[2]) / 3.f;
	}

	cellSize = 0.5f * avgObjectLength / scene.getObjects().size();
	invCellSize = 1.f / cellSize;

	deleteGrid();
	delete[] grid;

	cellsPerAxis = scene.getBounds().getLongestEdge() * invCellSize + 1;
	const unsigned N = cellsPerAxis * cellsPerAxis * cellsPerAxis;
	grid = new ObjectNode*[N];
	memset(grid, 0, N * sizeof(ObjectNode*));
#endif
}

void blurt::RegularGrid::build()
{
	deleteGrid();
	delete[] grid;

	// tests show this gives the best cellSize
	cellSize = 0.5f * avgObjectLength / scene.getObjects().size();
	invCellSize = 1.f / cellSize;
//	std::cout << "cellsize: " << cellSize << '\n';

	const Scene& s = Engine::get().getScene();
	sceneBounds = s.getBounds();
	pos = sceneBounds.pos1;
	float longestEdge = scene.getBounds().getLongestEdge();
	size = Vector3D(longestEdge, longestEdge, longestEdge),
	p2 = pos + size;
	cellsPerAxis = sceneBounds.getLongestEdge() * invCellSize + 1;
	const unsigned N = cellsPerAxis * cellsPerAxis * cellsPerAxis;
	grid = new ObjectNode*[N];
	memset(grid, 0, N * sizeof(ObjectNode*));

	float dx = cellSize;
	float dy = cellSize;
	float dz = cellSize;

	float dxInv = 1.0f/dx;
	float dyInv = 1.0f/dy;
	float dzInv = 1.0f/dz;

    avgObjectLength = 0.0f;
	foreach(const TraceableObject& o, scene.getObjects())
	{
		const AxisAlignedBoundingBox& aabb = o.getAxisAlignedBoundingBox();

		Vector3D diff1(aabb.pos1 - pos);
		Vector3D diff2(aabb.pos2 - pos);
		// find out which cells could contain the primitive (based on aabb)
		unsigned x1 = (unsigned) (diff1[0] * dxInv);
		unsigned x2 = (unsigned) (diff2[0] * dxInv) + 1;
		x1 = (x1 > cellsPerAxis) ? 0 : x1;
		x2 = (x2 > cellsPerAxis) ? cellsPerAxis : x2;

		unsigned y1 = (unsigned) (diff1[1] * dyInv);
		unsigned y2 = (unsigned) (diff2[1] * dyInv) + 1;
		y1 = (y1 > cellsPerAxis) ? 0 : y1;
		y2 = (y2 > cellsPerAxis) ? cellsPerAxis : y2;

		unsigned z1 = (unsigned) (diff1[2] * dzInv);
		unsigned z2 = (unsigned) (diff2[2] * dzInv) + 1;
		z1 = (z1 > cellsPerAxis) ? 0 : z1;
		z2 = (z2 > cellsPerAxis) ? cellsPerAxis : z2;


		Vector3D lower, higher; // corners of the current cell
		                        // (will be constructed throughout the loop)
		for (unsigned  x = x1; x < x2; ++x)
		{
			lower[0] = pos[0] + x * dx;
			higher[0] = lower[0] + dx;
			for (unsigned y = y1; y < y2; ++y)
			{
				lower[1] = pos[1] + y * dy;
				higher[1] = lower[1] + dy;
				for (unsigned z = z1; z < z2; ++z)
				{
					lower[2] = pos[2] + z * dz;
					higher[2] = lower[2] + dz;
					int idx = x + y * cellsPerAxis + z * cellsPerAxis * cellsPerAxis;

					// detailed intersection-test
					if (o.intersectsBox(lower, higher))
					{
						ObjectNode* n = new ObjectNode(o, grid[idx]);
						grid[idx] = n;
					}
				}
			}
		}
		const Vector3D size(aabb.pos2 - aabb.pos1);
		avgObjectLength += (size[0] + size[1] + size[2]) / 3.f;
	}
}


void blurt::RegularGrid::traceRay(const Ray& ray, IntersectionData* id) const
{
	PreparedRay pr;
	prepareRay(ray, &pr);
	if (!pr.valid)
		return;

	unsigned x = pr.startX;
	unsigned y = pr.startY;
	unsigned z = pr.startZ;
	Vector3D maxT(pr.maxT);

	id->reset();
	bool isIntersection = false;
	float currentDistance;
	while (true)
	{
		// check the current cell
		int idx = x + y * cellsPerAxis + z * cellsPerAxis * cellsPerAxis;
		ObjectNode* n = grid[idx];
		id->reset();
		while (n != 0)
		{
			isIntersection = n->obj.intersectsRay(ray, &currentDistance);
			id->testAndSetHit(&n->obj, currentDistance);
			n = n->next;
		}

		// advance to the next cell
		unsigned vecIdx; // in which direction do we advance
		if (maxT[0] < maxT[1])
		{
			if (maxT[0] < maxT[2])
			{
				x += pr.stepX;
				vecIdx = 0;
			}
			else
			{
				z += pr.stepZ;
				vecIdx = 2;
			}
		}
		else
		{
			if (maxT[1] < maxT[2])
			{
				y += pr.stepY;
				vecIdx = 1;
			}
			else
			{
				z += pr.stepZ;
				vecIdx = 2;
			}
		}

		// test if we found a valid Intersection
		if (id->getDistance() < maxT[vecIdx] + pr.distanceToGrid)
			return;
		maxT[vecIdx] += pr.deltaT[vecIdx];

		if (x >= cellsPerAxis || y >= cellsPerAxis || z >= cellsPerAxis)
			return;
	}
}


bool blurt::RegularGrid::isFirstIntersection(const Ray& ray, const float distance) const
{
	PreparedRay pr;
	prepareRay(ray, &pr);
	if (!pr.valid)
		return true;

	unsigned x = pr.startX;
	unsigned y = pr.startY;
	unsigned z = pr.startZ;
	Vector3D maxT(pr.maxT);

	bool isIntersection = false;
	float currentDistance;
	while (true)
	{
		// check the current cell
		unsigned idx = x + y * cellsPerAxis + z * cellsPerAxis * cellsPerAxis;
		ObjectNode* n = grid[idx];
		while (n != 0)
		{
			isIntersection = n->obj.intersectsRay(ray, &currentDistance);
			if (isIntersection && currentDistance < distance)
				return false;
			n = n->next;
		}

		// advance to the next cell
		unsigned vecIdx; // in which direction do we advance
		if (maxT[0] < maxT[1])
		{
			if (maxT[0] < maxT[2])
			{
				x += pr.stepX;
				vecIdx = 0;
			}
			else
			{
				z += pr.stepZ;
				vecIdx = 2;
			}
		}
		else
		{
			if (maxT[1] < maxT[2])
			{
				y += pr.stepY;
				vecIdx = 1;
			}
			else
			{
				z += pr.stepZ;
				vecIdx = 2;
			}
		}

		// check if we've already gone to far
		if (maxT[vecIdx] > distance || x >= cellsPerAxis
			|| y >= cellsPerAxis || z >= cellsPerAxis)
			return true;

		maxT[vecIdx] += pr.deltaT[vecIdx];
	}
}


void blurt::RegularGrid::prepareRay(const Ray& ray, PreparedRay* pr) const
{
	pr->valid = false;

	Vector3D origin(ray.getOrigin());

	// advance the ray to the grid
	pr->distanceToGrid = 0.f;
	if (!sceneBounds.contains(ray.getOrigin()))
	{
		pr->distanceToGrid = std::numeric_limits<float>::max();
		if (sceneBounds.intersectsRay(ray, &pr->distanceToGrid))
		{
			const Vector3D tmp = (pr->distanceToGrid + EPSILON) * ray.getDirection();
			origin += tmp;
		}
		else
		{
			pr->valid = false;
			return;
		}
	}

	Vector3D cell(origin - pos);
	cell *= invCellSize;

	// Cell-coordinates.
	// NOTE: we can never be in a "negative" cell, so using unsigned
	// allows us to only test for x > cellsPerAxis (w/o testing for x < 0)
	pr->startX = (unsigned) cell[0];
	pr->startY = (unsigned) cell[1];
	pr->startZ = (unsigned) cell[2];

	if ((cell[0] < 0.0f) || (pr->startX >= cellsPerAxis)
		|| (cell[1] < 0.0f) || (pr->startY  >= cellsPerAxis)
		|| (cell[2] < 0.0f) || (pr->startZ >= cellsPerAxis))
	{
		pr->valid = false;
		return;
	}

	const Vector3D& raydir = ray.getDirection();
	Vector3D raydirInv(1.0f / raydir[0], 1.0f / raydir[1], 1.0f / raydir[2]);

	// if raydir[x] should happen to be 0, we set maxT[x] to infinity,
	// so we will never make a step in that direction
	pr->stepX = raydir[0] > 0.f ? 1 : -1;
	pr->stepY = raydir[1] > 0.f ? 1 : -1;
	pr->stepZ = raydir[2] > 0.f ? 1 : -1;

	if (raydir[0] != 0.f)
	{
		float posNextCrossing = pos[0] + cellSize * (pr->startX + (pr->stepX > 0 ? 1 : 0));
		pr->maxT[0] = (posNextCrossing - origin[0]) * raydirInv[0];
		pr->deltaT[0] = (float)(cellSize * pr->stepX) * raydirInv[0];
	}
	else
		pr->maxT[0] = std::numeric_limits<float>::infinity();

	if (raydir[1] != 0.f)
	{
		float posNextCrossing = pos[1] + cellSize * (pr->startY + (pr->stepY > 0 ? 1 : 0));
		pr->maxT[1] = (posNextCrossing - origin[1]) * raydirInv[1];
		pr->deltaT[1] = (float)(cellSize * pr->stepY) * raydirInv[1];
	}
	else
		pr->maxT[1] = std::numeric_limits<float>::infinity();

	if (raydir[2] != 0.f)
	{
		float posNextCrossing = pos[2] + cellSize * (pr->startZ + (pr->stepZ > 0 ? 1 : 0));
		pr->maxT[2] = (posNextCrossing - origin[2]) * raydirInv[2];
		pr->deltaT[2] = (float)(cellSize * pr->stepZ) * raydirInv[2];
	}
	else
		pr->maxT[2] = std::numeric_limits<float>::infinity();

	pr->valid = true;
}
