#ifndef REGULARGRID_H
#define REGULARGRID_H

/**
 * RegularGrid.h
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "util/Vector3D.h"
#include "AxisAlignedBoundingBox.h"

namespace blurt
{

using util::Vector3D;
class TraceableObject;
class Scene;
class Ray;
class IntersectionData;


/**
 * The regular grid is a simple spacial subdivision scheme
 * to accelerate raytracing.
 */
class RegularGrid
{
	public:
		RegularGrid(Scene& scene);
		~RegularGrid();

		/// Rebuils the grid.
		void build();

		/**
		 * Optimizes the scene for the grid.
		 * Bigger triangles will be split in smaller once that
		 * better match the cellsize of the grid.
		 */
		void optimizeScene();

		/// Traces a ray trough the grid.
		void traceRay(const Ray& ray, IntersectionData* id) const;

		/**
		 * Returns true if the first intersection of the ray happens
		 * after the given distance, false otherwhise
		 */
		bool isFirstIntersection(const Ray& ray, float distance) const;

	private:
		void deleteGrid();

		// a simple linked list
		struct ObjectNode
		{
			const TraceableObject& obj;
			ObjectNode* next;

			ObjectNode(const TraceableObject& o, ObjectNode* n)
				: obj(o), next(n) {}
		};

		// stores the list of object in each grid-cell
		// (the 3D-structure has been flattened)
		ObjectNode** grid;

		Scene& scene;

		// position & size of the grid
		Vector3D pos;
		Vector3D size;

		// p2 = pos + size (we need this quite often,
		// so why not precalculate it ;)
		Vector3D p2;

		struct PreparedRay
		{
			Vector3D deltaT;
			Vector3D maxT;
			float distanceToGrid;
			unsigned startX;
			unsigned startY;
			unsigned startZ;
			int stepX;
			int stepY;
			int stepZ;
			bool valid;
		};

		void prepareRay(const Ray& ray,  PreparedRay* pr) const;
		float cellSize; // sidelength of a cell
		float invCellSize; // inverse cellsize
		unsigned int cellsPerAxis;

		AxisAlignedBoundingBox sceneBounds;

        // average sidelength of an objects AABB
        // (used to compute the new cellsize)
		float avgObjectLength;
};

};

#endif // REGULARGRID_H
