/**
 * Scene.cpp
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Scene.h"
#include "Plane.h"
#include "Triangle.h"
#include "Sphere.h"
#include "EngineError.h"
#include "AxisAlignedBoundingBox.h"
#include "ParticleSystem.h"
#include <fstream>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH


using namespace blurt;
using namespace util;


namespace
{
/*
 * creates the scene from Phantom's Raytracer-Tutorial on flipcode, Part 3
 */
#if 0
void createScene(Scene& s)
{
	s.setCameraPosition(Vector3D(0, 0, -5));
	s.setViewport(-4, -3, 8, 6);

	Material m1(Color( 0.7f, 0.7f, 1.0f ), 0.2, 0.8, 20, 0.2, 0.8, 1.3);
	s.addObject(new Sphere("big Sphere", m1, Vector3D(2, 0.8f, 3), 2.5));

	Material m2(Color( 0.7f, 0.7f, 1.0f), 0.1, 0.8, 20, 0.5, 0.0, 1.3);
	s.addObject(new Sphere("small Sphere", m2, Vector3D(-5.5f, -0.5, 7), 2));

	Material m5(Color( 1.0f, 0.4f, 0.4f), 0.2, 0.8, 20, 0.0, 0.8);
	s.addObject(new Sphere("extra Sphere", m5, Vector3D(-1.5, -3.8, 1), 1.5));

	Material lm1(Color(0.4f, 0.4f, 0.4f));
	s.addLight(new Sphere("Light 1", lm1, Vector3D(0, 5, 5), 0.1));

	Material lm2(Color(0.6f, 0.6f, 0.8f));
	s.addLight(new Sphere("Light 2", lm2, Vector3D(-3, 5, 1), 0.1));

	Material pm1(Color(0.4f, 0.3f, 0.3f), 1.0f, 0.8, 20, 0, 0);
	s.addObject(new Plane("ground Plane", pm1, Vector3D(0, 1, 0), 4.4));

	Material m6(Color( 0.5f, 0.3f, 0.5f), 0.6, 0.0);
	s.addObject(new Plane("back plane", m6, Vector3D(0.4f, 0, -1), 12));

	Material m7(Color( 0.4f, 0.7f, 0.7f ), 0.5f, 0.1f);
	s.addObject(new Plane("ceiling plane", m7, Vector3D(0, -1, 0), 7.4f));

	// grid
	std::string name = "Grid Sphere ";
	for ( int x = 0; x < 8; x++ ) for ( int y = 0; y < 7; y++ )
	{
		Material m(Color( 0.3f, 1.0f, 0.4f), 0.6, 0.6, 20, 0, 0);
		Vector3D pos(-4.5f + x * 1.5f, -1.3f + y * 1.5f, 10);
		char number = (char)(x * 8 + '0');
		s.addObject(new Sphere(name + number, m, pos, 0.3));
	}
}
#endif

#if 1
void createOrangeScene(Scene& s)
{
	s.setCameraPosition(Vector3D(0, 0, -5));
	s.setViewport(-4, -3, 8, 6);

	Material mat(Color( 1.0f, 153/256.0f, 0.0f ), 1, 0.5, 20, 0, 0, 1.3);
	Vector3D pos(0, 0, 2);
	for (int i = 0; i < 10; ++i)
	{
		Vector3D p(pos);
		int N = 400;
		p[0] += (rand() % N - N/2) / 100.0;
		p[1] += (rand() % N - N/2) / 100.0;
		p[2] += (rand() % N - N/2) / 100.0;
		float r = 0.25 + (rand() / (float) RAND_MAX);

		Color c (mat.getColor());
		N = 100;
		c[0] += (rand() % N - N/2) / 500.0;
		c[1] += (rand() % N - N/2) / 500.0;
		c[2] += (rand() % N - N/2) / 500.0;
		for (int i = 0; i < 3; ++i) c[i] = std::max(0.0f, std::min(1.0f, c[i]));

		float reflect = (rand() & 0x3) ? (rand() / float(1.6*RAND_MAX)) : 0.0f;
		float refract = (rand() & 0x7) ? (rand() / float(1.6*RAND_MAX)) : 0.0f;
		Material m(c, mat.getDiffuse(), mat.getSpecular(), 20, reflect, refract);
		s.addObject(new Sphere("sphere", m, p, r));
	}


	Material lm1(Color(1.0f, 1.0f, 1.0f));
	s.addLight(new Sphere("scene light", lm1, Vector3D(4, 4, 0.5), 0.1));


	Material lm2(Color(0.1f, 0.1f, 0.1f));
	s.addLight(new Sphere("scene light", lm1, Vector3D(0, 5, 0.5), 0.1));

//	Material lm2(Color(0.1f, 0.1f, 0.1f));
//	s.addLight(new Sphere("camera light", lm1, Vector3D(0, 0, -5), 0.1));

	Material pm1(Color(0.4f, 0.4f, 0.4f), 1.0f, 0, 20, 0, 0);
	s.addObject(new Plane("ground Plane", pm1, Vector3D(0, 1, 0), 4.4));
}
#endif

void createFractalSpheres(const Vector3D& up, const Vector3D& center,
                          float radius, int recursionDepth, Scene* s)
{
	const float RADIUS_DECREASE_FACTOR = 0.4f;
	const int NUMBER_OF_NEW_SPHERES = 6; // + 1 for the one on top ;)
	const Material SPHERE_MATERIAL(Color( 1, 0.7, 1), 0.7f);

	if (recursionDepth <= 0)
		return;

	// std::cout << "Center: " << center << " Up: " << up << '\n';

	s->addObject(new Sphere("fractalsphere", SPHERE_MATERIAL, center, radius));

	const float newRadius = RADIUS_DECREASE_FACTOR * radius;

/*
	// one sphere sits on top of the old one
	createFractalSpheres(up, center + (newRadius + radius) * up,
	                     newRadius, recursionDepth - 1, s);
*/

	const float alpha = (2.f * PI) / NUMBER_OF_NEW_SPHERES;

	// we put new spheres arround the current one in the plane defined by "up".
	// so we need to find a vector u perpendicular to up. That vector lies in the
	// plane, and u will be the new "up" for the next spheres. We'll then rotate
	// u in the scene.
	// we can freely choose 2 coordiniates of u, and calculate the third one the way
	// that it's perpendicular to up.
	Vector3D plane(1, 1, 1);
	if (up[2] != 0)
		plane[2] = (up[0] + up[1]) / up[2];
	else if (up[1] != 0)
		plane[1] = (up[0] + up[2]) / up[1];
	else
		plane = Vector3D(0, 1, 0);

	for (int i = 0; i < NUMBER_OF_NEW_SPHERES; ++i)
	{
		Vector3D u(plane);
		rotate(&u, up, alpha * i);
		u.normalize();

		createFractalSpheres(u, center + (radius + newRadius) * u,
	                     newRadius, recursionDepth - 1, s);
	}
}

#if 0
void createSphereflake(const Vector3D& center, float radius, int level, Scene* s)
{
	const float RADIUS_DECREASE_FACTOR = 0.5f;
	const Material SPHERE_MATERIAL(Color( 1, 0.7, 1), 0.7f);

	s->addObject(new Sphere("fractalsphere", SPHERE_MATERIAL, center, radius));

	if (level == 0)
		return;
	else
		--level;

	float rn = 3.0f * radius * invsqrt(12.0f);
	float r = radius * RADIUS_DECREASE_FACTOR;
	for (int dz = -1; dz <= 1; dz += 2)
	{
		for (int dx = -1; dx <= 1; dx += 2)
		{
			Vector3D v(dx, 1, dz);
			v.normalize();
			createSphereflake(center + (r + radius) * v, r, level, s);
		}
	}
}
#endif

/*
 * Splits big triangles into smaller ones.
 * Triangles will be split as long as the longest edge of their bounding
 * box is longer than minSize.
 * Note:
 * For Regular Grids, this actually doesn't give any speedup:
 * smaller triangles overlap more in adjacent cells,
 * so all in all we have (A LOT) more intersectiontests
 */
void splitBigTriangles(Scene* scene, float minSize)
{
	typedef Scene::ObjectContainer::iterator ObjIter;
	Scene::ObjectContainer& objects = scene->getObjects();

	ObjIter i = objects.begin();
	while (i != objects.end())
	{
		if (i->getType() != TraceableObject::TRIANGLE
			|| i->getAxisAlignedBoundingBox().getLongestEdge() <= minSize)
		{
			++i;
			continue;
		}

		Triangle& t = static_cast<Triangle&>(*i);


		std::string name = t.getName();
		const Material& m = t.getMaterial();

		// split the triangle by its longest axis
		const Vector3D& a = t.getVertex(0);
		const Vector3D& b = t.getVertex(1);
		const Vector3D& c = t.getVertex(2);

		Vector3D ab = a - b;
		float distAB = ab.getLength();

		Vector3D ac = a - c;
		float distAC = ac.getLength();

		Vector3D bc = b - c;
		float distBC = bc.getLength();

		Triangle* newTriangles[2];
		if ((distAB > distAC) && (distAB > distBC))
		{
			Vector3D splitPos =  b + 0.5f * ab;
			newTriangles[0] = new Triangle(name + "0", m, splitPos, c, a);
			newTriangles[1] = new Triangle(name + "1", m, splitPos, b, c);
		}
		else if (distBC > distAC)
		{
			Vector3D splitPos =  c + 0.5f * bc;
			newTriangles[0] = new Triangle(name + "0", m, splitPos, c, a);
			newTriangles[1] = new Triangle(name + "1", m, splitPos, a, b);
		}
		else
		{
			Vector3D splitPos =  c + 0.5f * ac;
			newTriangles[0] = new Triangle(name + "0", m, splitPos, a, b);
			newTriangles[1] = new Triangle(name + "1", m, splitPos, b, c);
		}

		i = objects.erase(i);
		i = objects.insert(i, newTriangles[0]);
		i = objects.insert(i, newTriangles[1]);
	}
}


/*
 * CONVERT PLANES INTO TRIANGLES
 * Basically we just find the 4 points of the plane at (x, y) = (+-limit, +-limit),
 * and then make 2 triangles out of it.
 * "Split big triangles" should take care of the rest :)
 *
 * Note that if the normal n of the plane has n.z == 0.0, we can't use this
 * method, so we'll set (x, z) = (+- limit, +- limit) and calculate y.
 */
void splitPlaneIntoTriangles(Scene* scene, float limit = 32)
{
	typedef Scene::ObjectContainer::iterator ObjIter;
	Scene::ObjectContainer& objects = scene->getObjects();

	ObjIter i = objects.begin();
	while (i != objects.end())
	{
		if (i->getType() != TraceableObject::PLANE)
		{
			++i;
			continue;
		}

		std::string name = i->getName();
		const Material& m = i->getMaterial();

		Plane& plane = static_cast<Plane&>(*i);
		const Vector3D& normal = plane.getNormal();
		float d = plane.getD();

		Vector3D p1(-limit, -limit, -limit);
		Vector3D p2(limit, -limit, -limit);
		Vector3D p3(-limit, limit, limit);
		Vector3D p4(limit, limit, limit);
		if (normal[2] != 0.0f)
		{
			p1[2] = (-p1[0]*normal[0] - p1[1]*normal[1] - d) / normal[2];
			p2[2] = (-p2[0]*normal[0] - p2[1]*normal[1] - d) / normal[2];
			p3[2] = (-p3[0]*normal[0] - p3[1]*normal[1] - d) / normal[2];
			p4[2] = (-p4[0]*normal[0] - p4[1]*normal[1] - d) / normal[2];
		}
		else if (normal[1] != 0.0f)
		{
			p1[1] = (-p1[0]*normal[0] - d) / normal[1];
			p2[1] = (-p2[0]*normal[0] - d) / normal[1];
			p3[1] = (-p3[0]*normal[0] - d) / normal[1];
			p4[1] = (-p4[0]*normal[0] - d) / normal[1];
		}
		else // normal == (+-1, 0, 0)
		{
			p1[0] = -d / normal[0];
			p2[0] = -d / normal[0];
			p3[0] = -d / normal[0];
			p4[0] = -d / normal[0];
		}

		Triangle* newTriangles[2];
		newTriangles[0] = new Triangle(name + "0", m, p3, p2, p1);
		newTriangles[1] = new Triangle(name + "1", m, p3, p4, p2);

		i = objects.erase(i);
		i = objects.insert(i, newTriangles[0]);
		i = objects.insert(i, newTriangles[1]);
	}
}

/*
 * adds a box (dscribed by lower and higher points) to the scene
 */
void addBox(Scene* scene, const Vector3D& lower, const Vector3D& higher)
{
	const Material m(Color( 0, 0, 1), 0.1f, 0, 0, 0, 1, 1);

	Vector3D v0(lower);
	Vector3D v1(lower);
	Vector3D v2(lower);

	v1[1] = higher[1];
	v2[0] = higher[0];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v0[0] = higher[0];
	v0[1] = higher[1];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v2 = higher;
	v2[0] = lower[0];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v1 = higher;
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v2 = higher;
	v2[1] = lower[1];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v1 = lower;
	v1[0] = higher[0];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v0 = lower;
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v1 = lower;
	v1[2] = higher[2];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v2 = lower;
	v2[1] = higher[1];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v0 = higher;
	v0[0] = lower[0];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v2 = higher;
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
	v0 = higher;
	v0[1] = lower[1];
	scene->addObject(new Triangle("box triangle", m, v0, v1, v2));
}

};


/*
 * NOTE:
 * we initialize the bounds with a very small value.
 * The "real" bounds will be determined once we read in the scene
 * from the file.
 * The Plane-Primitive returns the Scenes bounds as its own AABB,
 * so by setting this to some very small value we make sure that
 * the plane's AABB doesn't influence the calculation of these bounds
 * later on.
 */
blurt::Scene::Scene()
	: bounds(Vector3D(0, 0, 0), Vector3D(0, 0, 0))
{
	//ctor
}

blurt::Scene::~Scene()
{
	//dtor
}

void blurt::Scene::setCameraPosition(const Vector3D& pos)
{
	cameraPos = pos;
}


void blurt::Scene::setViewport(float x, float y, float width, float height)
{
	viewportX = x;
	viewportY = y;
	viewportWidth = width;
	viewportHeight = height;
}


void blurt::Scene::addObject(TraceableObject* obj)
{
	objects.push_back(obj);

	const AxisAlignedBoundingBox& aabb = obj->getAxisAlignedBoundingBox();
	if (bounds.pos1[0] > aabb.pos1[0]) bounds.pos1[0] = aabb.pos1[0];
	if (bounds.pos2[0] < aabb.pos2[0]) bounds.pos2[0] = aabb.pos2[0];
	if (bounds.pos1[1] > aabb.pos1[1]) bounds.pos1[1] = aabb.pos1[1];
	if (bounds.pos2[1] < aabb.pos2[1]) bounds.pos2[1] = aabb.pos2[1];
	if (bounds.pos1[2] > aabb.pos1[2]) bounds.pos1[2] = aabb.pos1[2];
	if (bounds.pos2[2] < aabb.pos2[2]) bounds.pos2[2] = aabb.pos2[2];
}


void blurt::Scene::removeObject(TraceableObject* obj)
{
	ObjectIter i = find(objects.begin(), objects.end(), *obj);
	if (i != objects.end())
		objects.erase(i);
	else
		std::cout << "w00t, could not delete that thing you gave me\n";
}


void blurt::Scene::addLight(Light* light)
{
	lights.push_back(light);
}


const blurt::TraceableObject* Scene::getObject(const std::string& name) const
{
#if TRACEABLEOBJECTS_HAVE_NAMES
	foreach (const TraceableObject& o, getObjects())
	{
		if (o.getName() == name)
		{
			return &o;
		}
	}
#endif

	return 0;
}


void blurt::Scene::update()
{
	recalculateBounds();
	foreach (ParticleSystem& s, particleSystems)
	{
		s.update();
	}
}


void blurt::Scene::recalculateBounds()
{
	float maxVal = std::numeric_limits<float>::max();
	Vector3D max(-maxVal, -maxVal, -maxVal);
	Vector3D min(maxVal, maxVal, maxVal);

	foreach (const TraceableObject& o, getObjects())
	{
		const AxisAlignedBoundingBox& aabb = o.getAxisAlignedBoundingBox();

		if (min[0] > aabb.pos1[0]) min[0] = aabb.pos1[0];
		if (max[0] < aabb.pos2[0]) max[0] = aabb.pos2[0];
		if (min[1] > aabb.pos1[1]) min[1] = aabb.pos1[1];
		if (max[1] < aabb.pos2[1]) max[1] = aabb.pos2[1];
		if (min[2] > aabb.pos1[2]) min[2] = aabb.pos1[2];
		if (max[2] < aabb.pos2[2]) max[2] = aabb.pos2[2];
	}

	for (size_t i = 0; i < 3; ++i)
	{
		if (min[i] == max[i])
		{
			min[i] -= 5.f;
			max[i] += 5.f;
		}
	}

	bounds =  AxisAlignedBoundingBox(min, max);
}


void blurt::Scene::writeToFile(const std::string& path) const
{
	std::ofstream f(path.c_str());
	if (!f)
		throw EngineError("failed to save scenery");

	f << "CAMERA: " << cameraPos << '\n';
	f << "VIEWPORT: x: " << viewportX
	  << "\ty: " << viewportY
	  << "\twidth: " << viewportWidth
	  << "\theight: " << viewportHeight << "\n\n";

	foreach (const TraceableObject& o, getObjects())
	{
		switch (o.getType())
		{
			case TraceableObject::SPHERE:
				f << "SPHERE: ";
				break;
			case TraceableObject::PLANE:
				f << "PLANE: ";
				break;
			case TraceableObject::TRIANGLE:
				f << "TRIANGLE: ";
				break;
		}
		f << o << "\n\n";
	}

	foreach (const Light& l, getLights())
	{
		f << "LIGHT: " << l << "\n\n";
	}

	foreach (const ParticleSystem& s, getParticleSystems())
	{
		f << "PARTICLESYSTEM:\n" << s << "\n\n";
	}

	f.close();
}


void blurt::Scene::readFromFile(const std::string& path)
{
	std::ifstream f(path.c_str());
	if (!f)
		throw EngineError("could not open scenery-file " + path);

	f.ignore(8);
	f >> cameraPos;
	f.ignore(14);
	f >> viewportX;
	f.ignore(3);
	f >> viewportY;
	f.ignore(7);
	f >> viewportWidth;
	f.ignore(8);
	f >> viewportHeight;
	f.ignore();

	// used to determine a good Boundaries-AABB
	float maxVal = std::numeric_limits<float>::max();
	Vector3D max(-maxVal, -maxVal, -maxVal);
	Vector3D min(maxVal, maxVal, maxVal);

	char tmp;
	while (f)
	{
		f >> tmp;

		TraceableObject* o = 0;
		switch (tmp)
		{
			case 'S': f.ignore(7); o = new Sphere(f); break;
			case 'P':  // could be a plane or a particle system
			{
				char p = f.peek();
				if (p == 'L')
				{
					f.ignore(6);
					o = new Plane(f);
				}
				else if (p == 'A')
				{
					f.ignore(15);
					particleSystems.push_back(new ParticleSystem(f));
				}
				break;
			}
			case 'T': f.ignore(9); o = new Triangle(f); break;
			case 'L': f.ignore(6); addLight(new Light(f)); break;
		}
		f.ignore(2); // newline

		if (o)
		{
			addObject(o);

			const AxisAlignedBoundingBox& aabb = o->getAxisAlignedBoundingBox();

			if (min[0] > aabb.pos1[0]) min[0] = aabb.pos1[0];
			if (max[0] < aabb.pos2[0]) max[0] = aabb.pos2[0];
			if (min[1] > aabb.pos1[1]) min[1] = aabb.pos1[1];
			if (max[1] < aabb.pos2[1]) max[1] = aabb.pos2[1];
			if (min[2] > aabb.pos1[2]) min[2] = aabb.pos1[2];
			if (max[2] < aabb.pos2[2]) max[2] = aabb.pos2[2];
			o = 0;
		}

		if(f.fail() && !f.eof())
		{
			f.close();
			throw EngineError("an error occured while processing " + path);
		}
	}
	f.close();

	for (size_t i = 0; i < 3; ++i)
	{
		if (min[i] == max[i])
		{
			min[i] -= 0.5f;
			max[i] += 0.5f;
		}
	}

	bounds = AxisAlignedBoundingBox(min, max);
//	addBox(this, Vector3D(-15, -15, 20), Vector3D(15, 15, 50));

/*
	this->objects.clear();
	this->lights.clear();
	createOrangeScene(*this);
*/
//	splitPlaneIntoTriangles(this, 15.f);
//	splitBigTriangles(this, 5);
	recalculateBounds();

//	createSphereflake(Vector3D(0, 0, 10), 3, 3, this);
//	createFractalSpheres(Vector3D(0, 0, 1), Vector3D(0, 0, 10), 3, 4, this);

//	particleSystems.push_back(new ParticleSystem(Vector3D(0, -2, 15), Vector3D(0, 1, 0)));
//	createScene(*this);
//	writeToFile("orange.txt");
}


