#include "Sphere.h"
#include "Ray.h"
#include "RayPacket.h"
#include "util/mathhelper.h"
#include "util/sse_helper.h"
#include <iostream>


using namespace util;
using namespace util::sse;

namespace blurt
{

Sphere::Sphere(const std::string& name,  const Material& material,
               const Vector3D& center_, float radius_)
	: TraceableObject(name, material),
	  center(center_),
	  radius(radius_),
	  radiusSquare(radius*radius)
{
}


Sphere::Sphere(std::istream& is)
{
	readFromFile(is);
}


Sphere::~Sphere()
{
	//dtor
}


bool Sphere::intersectsRayImpl(const Ray& ray, float* distance) const
{
	const Vector3D v(ray.getOrigin() - center);
	const float b = -dotProduct(v, ray.getDirection());
	float det = (b * b) - dotProduct(v, v) + radiusSquare;//radius*radius;
	if (det < 0.0f)
		return false;

	det = sqrtf(det);
	const float i1 = b - det;
	const float i2 = b + det;
	if (i2 > 0.0f)
	{
		*distance = (i1 < 0.0f) ? i2 : i1;
		return true;
	}
	return false;
}


void Sphere::intersectRayPacketImpl(RayPacket* rayPacket) const
{
	const VectorArraySSE& origins = rayPacket->getOrigins();
	const VectorArraySSE& directions = rayPacket->getDirections();

//	Vector3D v(ray.getOrigin() - center);
	const VectorArraySSE c(center[0], center[1],center[2]);
	const VectorArraySSE v(origins - c);

	// const float b = -dotProduct(v, ray.getDirection());
	const VectorSSE b = -dotProduct(v, directions);

	// float det = (b * b) - dotProduct(v, v) + radiusSquare;
	VectorSSE det = (b*b) - dotProduct(v, v) + VectorSSE(radiusSquare);

	// if (det < 0.0f)
	//	return MISS;
	//det = sqrtf(det);
	VectorSSE det_gtZeroMask(compareGt(det, ZERO_VEC));
	det &= det_gtZeroMask;
	det.sqrt();

	const VectorSSE i1((b - det) & det_gtZeroMask);
	const VectorSSE i2((b + det) & det_gtZeroMask);
	const VectorSSE i2_gtZeroMask = compareGt(i2, ZERO_VEC);

	if (i2_gtZeroMask.isZero())
		return;

	const VectorSSE i1_ltZeroMask = compareLt(i1, ZERO_VEC);
	const VectorSSE i1_gtZeroMask = compareGt(i1, ZERO_VEC);

	// where i2 > 0 && i1 < 0, the result is i2
	VectorSSE results(i2 & i2_gtZeroMask);
	results &= i1_ltZeroMask;
	results |= (i1 & i1_gtZeroMask);

	const VectorSSE validResults = det_gtZeroMask & i2_gtZeroMask;
	rayPacket->updateHitInformation(results, validResults, this);
}


bool Sphere::intersectsBoxImpl(const Vector3D& lower, const Vector3D& higher) const
{
	float minDistance = 0;

	for (size_t i = 0; i < 3; ++i)
	{
		if (center[i] < lower[i])
			minDistance += (center[i] - lower[i]) * (center[i] - lower[i]);
		else if (center[i] > higher[i])
			minDistance += (center[i] - higher[i]) * (center[i] - higher[i]);
	}

	return (minDistance <= radiusSquare); //radius * radius);
}


AxisAlignedBoundingBox Sphere::getAxisAlignedBoundingBoxImpl() const
{
	Vector3D radiusVec(radius, radius, radius);
	return AxisAlignedBoundingBox(center - radiusVec, center + radiusVec);
}


Vector3D blurt::Sphere::getNormalImpl(const Vector3D& pos) const
{
	 Vector3D retval(radius * (pos - center));
	 retval.normalize();
	 return retval;
}


void Sphere::writeToFile(std::ostream& os) const
{
	TraceableObject::writeToFile(os);
	os << '\n'
		<< "Center: " << center << "\tRadius: " << radius;
}


void Sphere::readFromFile(std::istream& is)
{
	TraceableObject::readFromFile(is);
	is.ignore(9);
	is >> center;
	is.ignore(9);
	is >> radius;
	radiusSquare = radius*radius;
}

}
