/**
 * TraceableObject.cpp
 *
 * (c) 2007 Thomas Unterthiner
 *
 */


#include "TraceableObject.h"
#include "RayPacket.h"
#include "IntersectionData.h"
#include <iostream>
#include <malloc.h>
#include <stdio.h>
#if defined(WIN32) && defined(__MINGW32__)
	#include <mm_malloc.h>
#endif

using namespace util;



namespace blurt
{

class Ray;

TraceableObject::TraceableObject(const std::string& name_, const Material& material_)
	: material(material_)
#if TRACEABLEOBJECTS_HAVE_NAMES
	  ,name(name_)
#endif
{
}

TraceableObject::TraceableObject()
{
}

TraceableObject::~TraceableObject()
{
	//dtor
}

void TraceableObject::writeToFile(std::ostream& os) const
{
#if TRACEABLEOBJECTS_HAVE_NAMES
	os << name << '\n' << material;
#else
	os << "traceable object" << '\n' << material;
#endif
}

void TraceableObject::readFromFile(std::istream& is)
{
#if TRACEABLEOBJECTS_HAVE_NAMES
	getline(is, name);
#else
	std::string tmp;
	getline(is, tmp);
#endif
	is >> material;
}


std::ostream& operator<< (std::ostream& os, const TraceableObject& o)
{
	o.writeToFile(os);
	return os;
}


std::istream& operator>> (std::istream& is, TraceableObject& o)
{
	o.readFromFile(is);
	return is;
}


/*
 * aligns on 16-Byte boundaries
 *
 * allocates more memory than needed. Stores the original pointer
 * in front of the memory that will be returned.
 */
void* TraceableObject::operator new(size_t size)
{
	const size_t alignment = 16;
	assert(0 < alignment && alignment <= 128);
	assert((alignment & (alignment - 1)) == 0);  // aligment must be a power of 2

/*
	// taken from the Ogre3D-Sourcecode
	unsigned char* p = new unsigned char[size + alignment];
	size_t offset = alignment - (size_t(p) & (alignment-1));

	unsigned char* result = p + offset;
	result[-1] = (unsigned char)offset;

	return result;
*/
	// for windows: http://msdn2.microsoft.com/en-US/library/8z34s9c6.aspx
	#ifdef WIN32
	return _mm_malloc(size, alignment);
	#else
	return memalign(alignment, size);
	#endif
}


void TraceableObject::operator delete(void* deadObject, size_t size)
{
	if (!deadObject)
		return;
/*
	// taken from the Ogre3D-Sourcecode
	unsigned char* mem = (unsigned char*)deadObject;
	mem = mem - mem[-1];
	delete [] mem;
*/
	#ifdef WIN32
	_mm_free(deadObject);
	#else
	free(deadObject);
	#endif
}

void TraceableObject::intersectRayPacketImpl(RayPacket* rayPacket) const
{
	float distance = 0.0f;
	bool isIntersection = false;
	for (int i = 0; i < 4; ++i)
	{
		distance = 0.0f;
		Ray ray = rayPacket->getRay(i);
		isIntersection = intersectsRay(ray, &distance);
		if (isIntersection)
			rayPacket->setIntersectionData(i, distance, this);
	}
}

}


