#ifndef TRACEABLEOBJECT_H
#define TRACEABLEOBJECT_H

/**
 * TraceableObject.h: baseclass for all the traceable primitives
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Material.h"
#include "util/Vector3D.h"
#include "AxisAlignedBoundingBox.h"
#include <iosfwd>
#include <string>


// NOTE: performance measurements indicate this has actually no impact
#define TRACEABLEOBJECTS_HAVE_NAMES 1

namespace blurt
{
using util::Vector3D;
class Ray;
class RayPacket;

class TraceableObject
{
	public:
		enum Type
		{
			SPHERE = 1,
			PLANE,
			TRIANGLE
		};

		TraceableObject();
		TraceableObject(const std::string& name, const Material& material);
		virtual ~TraceableObject();

		/**
		 * Tests if a ray intersects the object.
		 * Returns true or false. If there is an intersection, the distance
		 * from the intersection to the origin will be passed in by Reference
		 * in the distance parameter
		 */
		bool intersectsRay(const Ray& ray, float* distance) const
			{return intersectsRayImpl(ray, distance);}

		void intersectRayPacket(RayPacket* rayPacket) const
			{intersectRayPacketImpl(rayPacket);}

		/**
		 * Returns the normal of the object on position pos.
		 * It is assumed that pos is a point on the surface
		 * (if it isn't, the value returned is undefined!).
		 * The returned vector is already normalized!
		 */
		virtual Vector3D getNormal(const Vector3D& pos) const
			{return getNormalImpl(pos); }

		/**
		 * Returns an axis-aligned bounding box of the object.
		 */
		virtual AxisAlignedBoundingBox getAxisAlignedBoundingBox() const
			{return getAxisAlignedBoundingBoxImpl(); }


		/**
		 * Returns if the (axis-aligned) box described by its lower
		 * corner and the one diagonal to that intersects (or contains)
		 * the object.
		 */
		bool intersectsBox(const Vector3D& lower, const Vector3D& higher) const
			{return intersectsBoxImpl(lower, higher); }

		Type getType() const {return getTypeImpl(); }

		const Material& getMaterial() const  {return material;}

		#if TRACEABLEOBJECTS_HAVE_NAMES
		const ::std::string& getName() const {return name;}
		#else
		::std::string getName() const {return "noname";}
		#endif

		virtual void readFromFile(::std::istream& is);
		virtual void writeToFile(::std::ostream& os) const;

		// operator new is overloaded to make sure we have proper alignment
		static void* operator new(size_t size);
		static void operator delete(void* deadObject, size_t size);

	private:
		virtual bool intersectsRayImpl(const Ray& ray, float* distance) const = 0;
		virtual void intersectRayPacketImpl(RayPacket* rayPacket) const;
		virtual Vector3D getNormalImpl(const Vector3D& pos) const = 0;
		virtual AxisAlignedBoundingBox getAxisAlignedBoundingBoxImpl() const = 0;
		virtual bool intersectsBoxImpl(const Vector3D& lower, const Vector3D& higher) const = 0;
		virtual Type getTypeImpl() const = 0;

		Material material;
		#if TRACEABLEOBJECTS_HAVE_NAMES
		::std::string name; // for debug-reasons
		#endif
};

::std::ostream& operator<< (::std::ostream& os, const TraceableObject& o);
::std::istream& operator>> (::std::istream& is, TraceableObject& o);

inline bool operator==(const TraceableObject& a, const TraceableObject& b)
{
	return &a == &b;
}

};

#endif // TRACEABLEOBJECT_H
