/**
 * Triangle.cpp: a triangle-primitive
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

#include "Triangle.h"
#include "Ray.h"
#include "defaultsettings.h"
#include <iostream>

using namespace blurt;
using namespace util;



/*
 * if set to 1, the ray/triangle intersection algorithm will
 * be the one described in
 *	"Fast, minimum storage ray-triangle intersection."
 *	Tomas Mller and Ben Trumbore. Journal of Graphics Tools, 2(1):21--28, 1997
 *
 * Else, the one from Phantom's flipcode-tutorial will be used.
 * NOTE:
 * 	Phantom's method seemed to be better (by a factor of approx. 10/8), so we're
 *	using it :)
 */
#define USE_MOELLER_INTERSECTION 0

blurt::Triangle::Triangle(const std::string& name, const Material& material,
                  const Vector3D& v1, const Vector3D& v2, const Vector3D& v3)
	: TraceableObject(name, material)

{
	v[0] = v1;
	v[1] = v2;
	v[2] = v3;

	precalculateValues();
}



blurt::Triangle::~Triangle()
{
}



blurt::Triangle::Triangle(std::istream& is)
{
	readFromFile(is);
}


void blurt::Triangle::precalculateValues()
{
	const Vector3D v0(v[2] - v[0]);
	const Vector3D v1(v[1] - v[0]);
	normal = crossProduct(v0, v1);
	normal.normalize();
	planeD = -dotProduct(normal, v[0]);
//	std::cout << "plane d: " << planeD << " normal:" << normal << '\n';

	// precompute values for the intersection-test
	dot00 = dotProduct(v0, v0);
	dot01 = dotProduct(v0, v1);
	dot11 = dotProduct(v1, v1);
	invDenominator = 1.f / (dot00 * dot11 - dot01 * dot01);

}

void blurt::Triangle::readFromFile(std::istream& is)
{
	TraceableObject::readFromFile(is);
	for (size_t i = 0; i < 3; ++i)
	{
		is.ignore(5);
		is >> v[i];
	}

	precalculateValues();
}



void blurt::Triangle::writeToFile(std::ostream& os) const
{
	TraceableObject::writeToFile(os);
	os << '\n'
		<< "v1: " << v[0] << '\n'
		<< "v2: " << v[1] << '\n'
		<< "v3: " << v[2];
}



// taken from http://www.blackpawn.com/texts/pointinpoly/default.html
bool blurt::Triangle::isPointOnSurface(const Vector3D& point) const
{

	const Vector3D v0(v[2] - v[0]);
	const Vector3D v1(v[1] - v[0]);
	const Vector3D v2 = point - v[0];

	float dot02 = dotProduct(v0, v2);
	float dot12 = dotProduct(v1, v2);

	float u = (dot11 * dot02 - dot01 * dot12) * invDenominator;
	float v = (dot00 * dot12 - dot01 * dot02) * invDenominator;

	// Check if point is in triangle
	return (u > 0.f) && (v > 0.f) && (u + v < 1.f);
}



bool blurt::Triangle::intersectsRayImpl(const Ray& ray, float* distance) const
{
	#if !USE_MOELLER_INTERSECTION

	// Plane test
	float dot = dotProduct(normal, ray.getDirection());
	if (dot != 0.0f)
	{
		float dp = dotProduct(normal, ray.getOrigin());
		float dist = -(dp + planeD) / dot;
		if (dist > 0.0f)
		{
			// if we reached here, the ray intersects the plane after dist
			const Vector3D collPoint = ray.getOrigin() + ray.getDirection() * dist;
			if (isPointOnSurface(collPoint))
			{
				*distance = dist;
				return true;
			}
		}
	}
	return false;
#else
	// NOTE: the following 2 values could be precomputed (speedup by factor ~ 11/10)
	const Vector3D e1(v[1] - v[0]);
	const Vector3D e2(v[2] - v[0]);

	const Vector3D p(crossProduct(ray.getDirection(), e2));
	float det = dotProduct(e1, p);

	#if 0  // backface culling (does not work)
	if (det < EPSILON)
		return false;

	const Vector3D t (ray.getOrigin() - v[0]);
	float u = dotProduct(t, p);

	if (u < 0.f || u > det)
		return false;

	const Vector3D q(crossProduct(t, e1));

	float v = dotProduct(ray.getDirection(), q);
	if (v < 0.f || v > det)
		return false;

	*distance = dotProduct(e2, q);
	float invDet = 1.f / det;

	*distance *= invDet;
	u *= invDet;
	v *= invDet;
	#else
	if (det > -EPSILON && det < EPSILON)
		return false;

	float invDet = 1.f / det;
	const Vector3D t = ray.getOrigin() - v[0];
	float u = dotProduct(t, p) * invDet;

	if (u < 0.f || u > 1.f)
		return false;

	const Vector3D q(crossProduct(t, e1));
	float v = dotProduct(ray.getDirection(), q) * invDet;
	if (v < 0.f || u + v > 1.f)
		return false;

	*distance = dotProduct(e2, q) * invDet;
	#endif
	return true;

#endif
}


Vector3D blurt::Triangle::getNormalImpl(const Vector3D& pos) const
{
	return normal;
}


AxisAlignedBoundingBox blurt::Triangle::getAxisAlignedBoundingBoxImpl() const
{
	Vector3D min(v[0]);
	Vector3D max(v[0]);

	for (size_t i = 1; i < 3; ++i)
	{
		if (v[i][0] < min[0]) min[0] = v[i][0];
		if (v[i][1] < min[1]) min[1] = v[i][1];
		if (v[i][2] < min[2]) min[2] = v[i][2];

		if (v[i][0] > max[0]) max[0] = v[i][0];
		if (v[i][1] > max[1]) max[1] = v[i][1];
		if (v[i][2] > max[2]) max[2] = v[i][2];
	}


	// for triangles that are axis-aligned, else the AABB has a width of 0
	// on the aligned axis
	for (size_t i = 0; i < 3; ++i)
	{
		min[i] -= 0.1;
		max[i] += 0.1;
	}


	return AxisAlignedBoundingBox(min, max);
}



/*
 * the following code is all needed for the triangle/box intersection,
 * and is taken the implementation of the paper
 *	Fast 3D Triangle-Box Overlap Testing
 *	Tomas Akenine-Mller
 *	The graphicjournal of graphics tools, Volume 6, Number 1, 2001
 *
 * found at
 *	http://jgt.akpeters.com/papers/AkenineMoller01/tribox.html
 *
 */
namespace
{
#define FINDMINMAX(x0,x1,x2,min,max) \
  min = max = x0;   \
  if(x1<min) min=x1;\
  if(x1>max) max=x1;\
  if(x2<min) min=x2;\
  if(x2>max) max=x2;

bool doPlaneBoxOverlap(const Vector3D& normal,float d, const Vector3D& maxbox)
{
	Vector3D min, max;
	for (int q=0; q < 3; ++q)
	{
		if(normal[q] >= 0.0f)
		{
			min[q] = -maxbox[q];
			max[q] = maxbox[q];
		}
		else
		{
			min[q]= maxbox[q];
			max[q]= -maxbox[q];
		}
	}
	float side1 = dotProduct(normal, min) + d;
	if (side1 > 0.0f)
		return false;
	float side2 = dotProduct(normal, max) + d;
	if (side2 >= 0.0f)
		return true;

	return false;
}

#define X 0
#define Y 1
#define Z 2

/*======================== X-tests ========================*/
#define AXISTEST_X01(a, b, fa, fb)			   \
	p0 = a*v0[Y] - b*v0[Z];			       	   \
	p2 = a*v2[Y] - b*v2[Z];			       	   \
        if(p0<p2) {min=p0; max=p2;} else {min=p2; max=p0;} \
	rad = fa * boxhalfsize[Y] + fb * boxhalfsize[Z];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_X2(a, b, fa, fb)			   \
	p0 = a*v0[Y] - b*v0[Z];			           \
	p1 = a*v1[Y] - b*v1[Z];			       	   \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[Y] + fb * boxhalfsize[Z];   \
	if(min>rad || max<-rad) return 0;

/*======================== Y-tests ========================*/
#define AXISTEST_Y02(a, b, fa, fb)			   \
	p0 = -a*v0[X] + b*v0[Z];		      	   \
	p2 = -a*v2[X] + b*v2[Z];	       	       	   \
        if(p0<p2) {min=p0; max=p2;} else {min=p2; max=p0;} \
	rad = fa * boxhalfsize[X] + fb * boxhalfsize[Z];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_Y1(a, b, fa, fb)			   \
	p0 = -a*v0[X] + b*v0[Z];		      	   \
	p1 = -a*v1[X] + b*v1[Z];	     	       	   \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[X] + fb * boxhalfsize[Z];   \
	if(min>rad || max<-rad) return 0;

/*======================== Z-tests ========================*/

#define AXISTEST_Z12(a, b, fa, fb)			   \
	p1 = a*v1[X] - b*v1[Y];			           \
	p2 = a*v2[X] - b*v2[Y];			       	   \
        if(p2<p1) {min=p2; max=p1;} else {min=p1; max=p2;} \
	rad = fa * boxhalfsize[X] + fb * boxhalfsize[Y];   \
	if(min>rad || max<-rad) return 0;

#define AXISTEST_Z0(a, b, fa, fb)			   \
	p0 = a*v0[X] - b*v0[Y];				   \
	p1 = a*v1[X] - b*v1[Y];			           \
        if(p0<p1) {min=p0; max=p1;} else {min=p1; max=p0;} \
	rad = fa * boxhalfsize[X] + fb * boxhalfsize[Y];   \
	if(min>rad || max<-rad) return 0;
};

// ok, the actual intersection-method ;)
bool blurt::Triangle::intersectsBoxImpl(const Vector3D& lower, const Vector3D& higher) const
{
	const Vector3D boxhalfsize(0.5f * (higher - lower));
	const Vector3D boxcenter = lower + boxhalfsize;

	float min,max,p0,p1,p2,rad,fex,fey,fez;

	// move everything so that the boxcenter is in (0,0,0)
	const Vector3D v0(v[0] - boxcenter);
	const Vector3D v1(v[1] - boxcenter);
	const Vector3D v2(v[2] - boxcenter);

	const Vector3D e0(v1 - v0);
	const Vector3D e1(v2 - v1);
	const Vector3D e2(v0 - v2);

	fex = fabs(e0[X]);
	fey = fabs(e0[Y]);
	fez = fabs(e0[Z]);
	AXISTEST_X01(e0[Z], e0[Y], fez, fey);
	AXISTEST_Y02(e0[Z], e0[X], fez, fex);
	AXISTEST_Z12(e0[Y], e0[X], fey, fex);

	fex = fabs(e1[X]);
	fey = fabs(e1[Y]);
	fez = fabs(e1[Z]);
	AXISTEST_X01(e1[Z], e1[Y], fez, fey);
	AXISTEST_Y02(e1[Z], e1[X], fez, fex);
	AXISTEST_Z0(e1[Y], e1[X], fey, fex);

	fex = fabs(e2[X]);
	fey = fabs(e2[Y]);
	fez = fabs(e2[Z]);
	AXISTEST_X2(e2[Z], e2[Y], fez, fey);
	AXISTEST_Y1(e2[Z], e2[X], fez, fex);
	AXISTEST_Z12(e2[Y], e2[X], fey, fex);

	/* test in X-direction */
	FINDMINMAX(v0[X],v1[X],v2[X],min,max);
	if(min>boxhalfsize[X] || max<-boxhalfsize[X]) return false;

	/* test in Y-direction */
	FINDMINMAX(v0[Y],v1[Y],v2[Y],min,max);
	if(min>boxhalfsize[Y] || max<-boxhalfsize[Y]) return false;

	/* test in Z-direction */
	FINDMINMAX(v0[Z],v1[Z],v2[Z],min,max);
	if(min>boxhalfsize[Z] || max<-boxhalfsize[Z]) return false;

	// compute new normal and d in respect to the new coordinates
	const Vector3D normal (crossProduct(e0, e1));
	float d = -dotProduct(normal, v0);
	if(!doPlaneBoxOverlap(normal, d, boxhalfsize)) return false;

	return true;
}


