/**
 * Framework_sdl.cpp:
 *	SDL-centric implementation of the framework.
 *
 * (c) 2007 Thomas Unterthiner
 *
 */


#include "Framework.h"
#include "Engine.h"
#include "blurt/EngineError.h"
#include "util/Color.h"
#include <SDL/SDL.h>

using util::Color;
using blurt::EngineError;

namespace
{
inline void putPixel(int x, int y, SDL_Surface* screen, const Color& c)
{
	y = screen->h - 1 - y; // rotate coordinates by their y values

	int bpp = screen->format->BytesPerPixel;
	Uint8* p = (Uint8*)screen->pixels + y * screen->pitch + x * bpp;

	// convert Color value to SDL-color
	int r = static_cast<int>(255 * c[0] + 0.5);
	int g = static_cast<int>(255 * c[1] + 0.5);
	int b = static_cast<int>(255 * c[2] + 0.5);
	if (r > 255) r = 255;
	if (g > 255) g = 255;
	if (b > 255) b = 255;
	Uint32 pixel_value = SDL_MapRGB(screen->format, r, g, b);

	switch(bpp)
	{
		case 1:
			*p = pixel_value;
			break;
		case 2:
			*(Uint16*)p = pixel_value;
			break;
		case 3:
			#if SDL_BYTEORDER == SDL_LIL_ENDIAN
				p[0] = pixel_value & 0xff;
				p[1] = (pixel_value >> 8) & 0xff;
				p[2] = (pixel_value >> 16) & 0xff;
			#else
				p[0] = (pixel_value >> 16) & 0xff;
				p[1] = (pixel_value >> 8) & 0xff;
				p[2] = pixel_value & 0xff;
			#endif
			break;
		case 4:
			*(Uint32*)p = pixel_value;
			break;
	}
}
}


namespace platform
{

/**
 * Intializes the framework.
 */
void init()
{
	if (SDL_Init(0) < 0)
		throw EngineError("unable to initialize SDL", SDL_GetError());
	atexit(SDL_Quit);

	// disable all events we don't care about
	for (int i = 0; i < SDL_NUMEVENTS; ++i)
	{
		if (i != SDL_KEYDOWN && i != SDL_QUIT)
			SDL_EventState(i, SDL_IGNORE);
	}
}


/**
 * Returns some kind of time measurement.
 */
unsigned int  getTicks()
{
	return SDL_GetTicks();
}


/**
 * Deals with all kind of input.
 */
bool handleInput(const Engine& engine)
{
	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
			case SDL_QUIT:
				return false;
				break;
/*
			case SDL_VIDEORESIZE:
				windowWidth = event.resize.w;
				windowHeight = event.resize.h;
				Renderer.resize(windowWidth, windowHeight);
				break;
*/
			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_ESCAPE)
					return false;
				else if (event.key.keysym.sym == SDLK_s)
					engine.saveScene("current_scene.txt");
				break;
		}
	}
	return true;
}


void sleep(unsigned int milliseconds)
{
	SDL_Delay(milliseconds);
}


struct Screen::ScreenImpl
{
	SDL_Surface* screen;
};

Screen::Screen(int screenWidth, int screenHeight, const char* windowTitle)
	: d(new ScreenImpl())
{
	if (!(SDL_WasInit(SDL_INIT_VIDEO) & SDL_INIT_VIDEO))
	{
		if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0)
			throw EngineError("unable to initialize the video subsystem",
				SDL_GetError());
	}

	unsigned flags = SDL_HWSURFACE;
	d->screen = SDL_SetVideoMode(screenWidth, screenHeight, 0, flags);
	if (d->screen == 0)
		throw EngineError("unable to set the desired video-mode", SDL_GetError());
	SDL_WM_SetCaption(windowTitle, 0);
}



Screen::~Screen()
{
	delete d;
}


void Screen::draw(const Color* image)
{
#if 1
	int w = getWidth();
	int h = getHeight();

	if (SDL_MUSTLOCK(d->screen))
		SDL_LockSurface(d->screen);

	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			putPixel(x, y, d->screen, image[y * w + x]);
		}
	}

	if (SDL_MUSTLOCK(d->screen))
		SDL_UnlockSurface(d->screen);
#endif
}


void Screen::save(const char* path)
{
	int retval = SDL_SaveBMP(d->screen, path);
	if (retval != 0)
		throw EngineError("Unable to save file", SDL_GetError());
}


int Screen::getWidth() const
{
	return d->screen->w;
}


int Screen::getHeight() const
{
	return d->screen->h;
}

void Screen::update() const
{
	SDL_Flip(d->screen);
};


void shutdown(platform::Screen& s)
{
}
};



