/**
 * Framework_sdl.cpp:
 *	SDL-centric implementation of the framework.
 *
 * (c) 2007 Thomas Unterthiner
 *
 */


#include "Framework.h"
#include "Engine.h"
#include "blurt/EngineError.h"
#include "util/Color.h"
#include <SDL/SDL.h>
#include <GL/gl.h>

using blurt::EngineError;

namespace platform
{

/**
 * Intializes the framework.
 */
void init()
{
	if (SDL_Init(0) < 0)
		throw EngineError("unable to initialize SDL", SDL_GetError());
	atexit(SDL_Quit);

	// disable all events we don't care about
	for (int i = 0; i < SDL_NUMEVENTS; ++i)
	{
		if (i != SDL_KEYDOWN && i != SDL_QUIT)
			SDL_EventState(i, SDL_IGNORE);
	}
}


/**
 * Returns some kind of time measurement.
 */
unsigned int  getTicks()
{
	return SDL_GetTicks();
}


/**
 * Deals with all kind of input.
 */
bool handleInput(const Engine& engine)
{
	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
			case SDL_QUIT:
				return false;
				break;
/*
			case SDL_VIDEORESIZE:
				windowWidth = event.resize.w;
				windowHeight = event.resize.h;
				Renderer.resize(windowWidth, windowHeight);
				break;
*/
			case SDL_KEYDOWN:
				if (event.key.keysym.sym == SDLK_ESCAPE)
					return false;
				else if (event.key.keysym.sym == SDLK_s)
					engine.saveScene("current_scene.txt");
				break;
		}
	}
	return true;
}


void sleep(unsigned int milliseconds)
{
	SDL_Delay(milliseconds);
}

struct Screen::ScreenImpl
{
	SDL_Surface* screen;
};


Screen::Screen(int screenWidth, int screenHeight, const char* windowTitle)
	: d(new ScreenImpl())
{
	if (!(SDL_WasInit(SDL_INIT_VIDEO) & SDL_INIT_VIDEO))
	{
		if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0)
			throw EngineError("unable to initialize the video subsystem",
				SDL_GetError());
	}

	d->screen = SDL_SetVideoMode(screenWidth, screenHeight, 0, SDL_OPENGL);
	if (d->screen == 0)
		throw EngineError("unable to set the desired video-mode", SDL_GetError());
	SDL_WM_SetCaption(windowTitle, 0);

	glViewport (0, 0, screenWidth, screenHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, screenWidth, 0, screenHeight, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}


Screen::~Screen()
{
    delete d;
}


void Screen::draw(const Color* image)
{
	static int w = getWidth();
	static int h = getHeight();

	glBegin(GL_POINTS);
	glClear(GL_COLOR_BUFFER_BIT);
	for (int y = 0; y < h; ++y)
	{
		const int offset = y * w;
		for (int x = 0; x < w; ++x)
		{
			const Color& c = image[offset + x];
			glColor3fv(c.getAddress(0));
			glVertex2i(x, y);
		}
	}
	glEnd();
}


void Screen::save(const char* path)
{
	int retval = SDL_SaveBMP(d->screen, path);
	if (retval != 0)
		throw EngineError("Unable to save file", SDL_GetError());
}


int Screen::getWidth() const
{
	return d->screen->w;
}


int Screen::getHeight() const
{
	return d->screen->h;
}

void Screen::update() const
{
	SDL_GL_SwapBuffers();
};

void shutdown(platform::Screen& s)
{
}

};


