/**
 * TheoraEncoder.cpp: a simple theora encoder implmemented using libtheora
 *
 * NOTE: mostly derived from
 * http://www.quinthar.com/index.php?page=ProgrammingWithTheora
 *
 * (c) 2007 Thomas Unterthiner
 *
 */


#include "TheoraEncoder.h"
#include <fstream>
#include <theora/theora.h>
#include <cstdlib>
#include "blurt/EngineError.h"
#include <boost/scoped_array.hpp>

using namespace std;
using blurt::EngineError;

const int BITRATE = 2000000; // Anywhere between 45kbps and 2000kbps
typedef unsigned char byte;


#define SHOW_YUV_IMAGE_ON_SCREEN 0
#if SHOW_YUV_IMAGE_ON_SCREEN
	#include <SDL/SDL.h>
#endif

namespace
{

/*
 * A simple box-filter for u- and v-values
 */
void boxFilter(byte* uIn, byte* vIn,byte* uOut, byte* vOut,
               unsigned width, unsigned height)
{
	// ok, now let the subsampling begin \o/
	size_t uvIdx = 0;
	for (size_t y = 0; y < height; y+= 2)
	{
		for (size_t x = 0; x < width; x += 2)
		{
			size_t row = y * width;
			size_t row1 = (y+1) * width;

			float uSum = uIn[row + x] + uIn[row + x + 1]
						+ uIn[row1 + x] + uIn[row1 + x + 1];
			unsigned uVal = uSum / 4;
			if (uVal > 255) uVal = 255;

			unsigned vSum = vIn[row + x + 1] + vIn[row + x + 2]
						+ vIn[row1 + x + 1] + vIn[row1 + x + 2];
			unsigned vVal = vSum / 4;
			if (vVal > 255) vVal = 255;

			uOut[uvIdx] = (byte) uVal;
			vOut[uvIdx++] = (byte) vVal;
		}
	}
}
}

namespace util
{

class TheoraEncoder::TheoraEncoderImpl
{
	public:
		TheoraEncoderImpl(const string& outputfile, unsigned width, unsigned height, unsigned fps);
		~TheoraEncoderImpl();
		void addFrame(unsigned width, unsigned height, Color* image, bool isLastFrame);

	private:
		void convertToYUV420p(unsigned width, unsigned height, Color* image) const;
		fstream fout;
		theora_info encoderInfo;
		theora_state theoraState;
		ogg_stream_state oggState;
		ogg_page page;
		ogg_packet packet;
		byte* yBuffer;
		byte* uBuffer;
		byte* vBuffer;
};


TheoraEncoder::TheoraEncoderImpl::~TheoraEncoderImpl()
{
	ogg_stream_clear(&oggState);
	theora_clear(&theoraState);
	fout.close();
	delete[] yBuffer;
	delete[] uBuffer;

	delete[] vBuffer;
}


TheoraEncoder::TheoraEncoderImpl::TheoraEncoderImpl(const string& outputfile,
                                     unsigned width, unsigned height, unsigned fps)
	: fout(outputfile.c_str(),  ios::out | ios::binary)
{
	if (!fout)
		throw EngineError("unable to open video outputfile: " + outputfile);

	if (width % 16 != 0 || height % 16 != 0)
		throw EngineError("widht and height must be multiples of 16!");

	if (ogg_stream_init(&oggState,rand()))
		throw EngineError("failed to initialize the ogg stream");

	theora_info_init( &encoderInfo );
	encoderInfo.frame_width = width;
	encoderInfo.frame_height = height;
	encoderInfo.width = encoderInfo.frame_width;
	encoderInfo.height = encoderInfo.frame_height;
	encoderInfo.offset_x = 0;
	encoderInfo.offset_y = 0;
	encoderInfo.fps_numerator  = fps;
	encoderInfo.fps_denominator = 1;
	encoderInfo.aspect_numerator = 1;
	encoderInfo.aspect_denominator = 1;
	encoderInfo.colorspace  = OC_CS_UNSPECIFIED;
	encoderInfo.pixelformat = OC_PF_420;
	encoderInfo.target_bitrate = BITRATE;
	encoderInfo.quality = 16;
	encoderInfo.dropframes_p = 0;
	encoderInfo.quick_p = 1;
	encoderInfo.keyframe_auto_p = 1;
	encoderInfo.keyframe_frequency = 64;
	encoderInfo.keyframe_frequency_force = 64;
	encoderInfo.keyframe_data_target_bitrate = encoderInfo.target_bitrate * 3 / 2;
	encoderInfo.keyframe_auto_threshold = 80;
	encoderInfo.keyframe_mindistance = 8;
	encoderInfo.noise_sensitivity = 1;

	if (theora_encode_init( &theoraState, &encoderInfo ))
		throw EngineError("failed to init Theora encoder");
	theora_info_clear( &encoderInfo );


	ogg_packet headerPacket;
	if (theora_encode_header( &theoraState, &headerPacket))
		throw EngineError("failed to encode Theora header");

  ogg_stream_packetin(&oggState,&headerPacket);
  if(ogg_stream_pageout(&oggState, &page) !=1 )
		throw EngineError("Internal Ogg library error.");

	fout.write((const char*)page.header, page.header_len);
	fout.write((const char*)page.body, page.body_len);

	theora_comment encoderComment;
	theora_comment_init( &encoderComment );
	if (theora_encode_comment(&encoderComment, &packet))
		throw EngineError("failed to encode Theora comment");

	ogg_stream_packetin(&oggState,&packet);
	theora_encode_tables( &theoraState, &packet);
	ogg_stream_packetin(&oggState,&packet);

	theora_comment_clear( &encoderComment );

	// we have tu flush after the header :)
	while (1)
	{
		int result = ogg_stream_flush(&oggState, &page);

		if (result < 0)
		{
			throw EngineError("Internal Ogg library error.");
			exit(1);
		}

		if (result == 0) break;

		fout.write((const char*)page.header, page.header_len);
		fout.write((const char*)page.body, page.body_len);
	}

	yBuffer = new byte[width*height];
	uBuffer = new byte[width*height / 4];
	vBuffer = new byte[width*height / 4];
}


/*
 * fills the y, u and v buffers.
 * NOTE: it doesn't seem to make a difference (performance-wise)
 *       if we do this in single or double-precision, so we use double
 */
void TheoraEncoder::TheoraEncoderImpl::convertToYUV420p(unsigned width,
				unsigned height, Color* image) const
{
	// we store all u and v values and subsample later on
	boost::scoped_array<byte> uVal(new byte[width * height]);
	boost::scoped_array<byte> vVal(new byte[width * height]);

	size_t idx = 0;
	// we invert the y-position by scaning from bottom to top,
	// because theora expects it that way :)
	for (int yPos = height - 1;  yPos >= 0; --yPos)
	{
		for (unsigned xPos = 0; xPos < width; ++xPos)
		{
			size_t pos =  yPos * width + xPos;
			double r = image[pos][0] > 1.0f ? 1.0 : image[pos][0];
			double g = image[pos][1] > 1.0f ? 1.0 : image[pos][1];
			double b = image[pos][2] > 1.0f ? 1.0 : image[pos][2];

			byte y = 66.0 * r + 129.0 * g + 25.0 * b + 16.0;
			byte u = -38.0 * r - 74.0 * g + 112.0 * b + 128.0;
			byte v = 112.0 * r - 94.0 * g - 18.0 * b + 128.0;

			// store u and v for the subsampling later on
			uVal[idx] = u;
			vVal[idx] = v;
			yBuffer[idx++] = y;
		}
	}

	boxFilter(uVal.get(), vVal.get(), uBuffer, vBuffer, width, height);
}


void TheoraEncoder::TheoraEncoderImpl::addFrame(unsigned width, unsigned height,
                                                 Color* image, bool isLastFrame)
{
	convertToYUV420p(width, height, image);

	// A little test: display the Image on the screen
#if SHOW_YUV_IMAGE_ON_SCREEN
	SDL_Surface* s = SDL_GetVideoSurface();
	if (s)
	{
		SDL_Overlay* ol = SDL_CreateYUVOverlay(s->w, s->h, SDL_IYUV_OVERLAY, s);
		SDL_LockYUVOverlay(ol);
		for (unsigned i = 0; i < height; ++i)
		{
			unsigned a = i * ol->pitches[0];
			unsigned b = i * ol->pitches[1];
			memcpy(ol->pixels[0] + a, yBuffer + a, ol->pitches[0]);
			if (i < height / 2)
			{
				memcpy(ol->pixels[1] + b, uBuffer + b, ol->pitches[1]);
				memcpy(ol->pixels[2] + b, vBuffer + b, ol->pitches[2]);
			}
		}
		SDL_UnlockYUVOverlay(ol);
		SDL_Rect r = {0, 0, s->w, s->h};
		SDL_DisplayYUVOverlay(ol, &r);
		SDL_FreeYUVOverlay(ol);
	}
#endif

	yuv_buffer yuv;
	yuv.y_width = width;
	yuv.y_height = height;
	yuv.y_stride = width;
	yuv.y  = yBuffer;
	yuv.uv_width = width / 2;
	yuv.uv_height = height / 2;
	yuv.uv_stride = yuv.uv_width;
	yuv.u = uBuffer;
	yuv.v = vBuffer;

	int retval = theora_encode_YUVin(&theoraState, &yuv);

	if (retval)
		throw EngineError("unable to encode theora packet (1)");

	theora_encode_packetout(&theoraState, isLastFrame, &packet);
	ogg_stream_packetin(&oggState, &packet);

	while (ogg_stream_pageout(&oggState, &page) != 0)
	{
		fout.write((const char*)page.header, page.header_len);
		fout.write((const char*)page.body, page.body_len);
	}
}



TheoraEncoder::TheoraEncoder(const std::string& outputfile,int width,
                             int height, int fps)
	: d(new TheoraEncoderImpl(outputfile, width, height, fps))
{
}



TheoraEncoder::~TheoraEncoder()
{
	delete d;
}



void TheoraEncoder::addFrame(int width, int height, Color* image, bool isLastFrame)
{
	d->addFrame(width, height, image, isLastFrame);
}

};
