
#ifdef VECTOR3D_H

#include "Vector3D.h"
#include <iostream>

/**
 * Vector3D.cpp
 * 	NOTE: to be used as an include-file so that everything
 *        in here is inline
 *
 * (c) 2007 Thomas Unterthiner
 *
 */

using namespace util;

inline util::Vector3D::Vector3D()
{
	v[0] = 0.f;
	v[1] = 0.f;
	v[2] = 0.f;
	v[3] = 0.f;
}

inline util::Vector3D::Vector3D(float x, float y, float z)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
	v[3] = 0.f;
}

inline util::Vector3D::Vector3D(const Vector3D& other)
{
	v[0] = other.v[0];
	v[1] = other.v[1];
	v[2] = other.v[2];
	v[3] = other.v[3];
}


inline Vector3D& util::Vector3D::operator-()
{
	*this *= -1;
	return *this;
}

inline Vector3D& util::Vector3D::operator+= (const Vector3D& other)
{
#if USE_GCC_VECTORISATION
	vec += other.vec;
#else
	v[0] += other[0];
	v[1] += other[1];
	v[2] += other[2];
#endif
	return *this;
}


inline Vector3D& util::Vector3D::operator*= (float a)
{
	v[0] *= a;
	v[1] *= a;
	v[2] *= a;
	return *this;
}

inline Vector3D& util::Vector3D::operator/= (float a)
{
	v[0] /= a;
	v[1] /= a;
	v[2] /= a;
	return *this;
}


inline bool util::Vector3D::operator==(const Vector3D& other) const
{
#if USE_GCC_VECTORISATION
	v4si compare = __builtin_ia32_cmpeqps(vec, other.vec);
	return __builtin_ia32_movmskps((v4sf)compare);
#else
	return v[0] == other[0]
		&& v[1] == other[1]
		&& v[2] == other[2];
#endif
}


inline bool util::Vector3D::operator!=(const Vector3D& other) const
{
	return !(*this == other);
}


inline Vector3D& util::Vector3D::operator-= (const Vector3D& other)
{
#if USE_GCC_VECTORISATION
	vec -= other.vec;
#else
	v[0] -= other[0];
	v[1] -= other[1];
	v[2] -= other[2];
#endif
	return *this;
}


inline const Vector3D util::operator+(const Vector3D& lhs, const Vector3D& rhs)
{
#if USE_GCC_VECTORISATION
	return Vector3D(lhs.vec + rhs.vec);
#else
	Vector3D retval(lhs);
	retval += rhs;
	return retval;
#endif
}


inline const Vector3D util::operator-(const Vector3D& lhs, const Vector3D& rhs)
{
#if USE_GCC_VECTORISATION
	return Vector3D(lhs.vec - rhs.vec);
#else
	Vector3D retval(lhs);
	retval -= rhs;
	return retval;
#endif
}


inline const Vector3D util::operator*(const Vector3D& lhs, const Vector3D& rhs)
{
#if USE_GCC_VECTORISATION
	return Vector3D(lhs.vec * rhs.vec);
#else
	Vector3D retval(lhs);
	retval[0] *= rhs[0];
	retval[1] *= rhs[1];
	retval[2] *= rhs[2];
	return retval;
#endif
}


inline const Vector3D util::operator*(float lhs, const Vector3D& rhs)
{
	Vector3D retval(rhs);
	retval *= lhs;
	return retval;
}


inline const Vector3D util::operator*(const Vector3D& lhs, float rhs)
{
	return rhs * lhs;
}


inline const Vector3D util::operator/(const Vector3D& lhs, float rhs)
{
	Vector3D retval(lhs);
	retval /= rhs;
	return retval;
}



inline Vector3D util::crossProduct(const Vector3D& lhs, const Vector3D& rhs)
{
		const float x = lhs[1] * rhs[2] - lhs[2] * rhs[1];
		const float y = lhs[2] * rhs[0] - lhs[0] * rhs[2];
		const float z = lhs[0] * rhs[1] - lhs[1] * rhs[0];
		return Vector3D(x, y, z);
}


inline float util::dotProduct(const Vector3D& lhs, const Vector3D& rhs)
{
	return lhs[0] * rhs[0] + lhs[1] * rhs[1] + lhs[2] * rhs[2];
}


inline void util::Vector3D::normalize()
{
	float isq = invsqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
	*this *= isq;
}

inline float util::Vector3D::getLength() const
{
	float tmp = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
#if USE_FAST_INVSQRT
	return 1.0f / invsqrt(tmp);
#else
	return sqrtf(tmp);
#endif
}



inline std::ostream& util::operator<< (std::ostream& stream, const Vector3D& v)
{
	stream << '(' << v[0] << ", " << v[1] << ", " << v[2] << ')';
	return stream;
}


inline std::istream& util::operator>> (std::istream& is, Vector3D& v)
{
	is.ignore();
	is >> v.v[0];
	is.ignore();
	is >> v.v[1];
	is.ignore();
	is >> v.v[2];
	is.ignore();
	return is;
}

/// Rotates the Vector arround the axis defined by the origin and the given axis
inline void util::rotate(Vector3D* v, const Vector3D& axis, float alpha)
{
	// formulas were derived by Mathematica, so sue me ;p
	float a = axis[0];
	float b = axis[1];
	float c = axis[2];

	float x = (*v)[0];
	float y = (*v)[1];
	float z = (*v)[2];

	const float l = a*a + b*b + c*c;
	const float invL = 1.0f / l;
	const float sqrtL = sqrt(l);
	const float sinAlpha = sin(alpha);
	const float cosAlpha = cos(alpha);
	const float dp = dotProduct(*v, axis);

	(*v)[0] = (a*dp + ( b*b*x - a*b*y + c*( c*x - a*z))*cosAlpha + sqrtL*(-c*y + b*z)*sinAlpha) * invL;
	(*v)[1] = (b*dp + (-a*b*x + a*a*y + c*( c*y - b*z))*cosAlpha + sqrtL*( c*x - a*z)*sinAlpha) * invL;
	(*v)[2] = (c*dp + (-a*c*x + a*a*z + b*(-c*y + b*z))*cosAlpha + sqrtL*(-b*x + a*y)*sinAlpha) * invL;
}

#endif
