#ifdef VECTORARRAYSSE_H
/**
 * VectorArraySSE.cpp
 * 	NOTE: to be used as an include-file so that everything
 *        in here is inline
 *
 * (c) 2007, 2008 Thomas Unterthiner
 *
 */

#include "VectorArraySSE.h"
#include "util/Vector3D.h"
#include <xmmintrin.h>
#include <emmintrin.h>
#include <pmmintrin.h>

using util::Vector3D;

namespace util
{
namespace sse
{

extern const VectorSSE HALF_VEC;
extern const VectorSSE THREE_VEC;

inline const VectorArraySSE operator*(const VectorArraySSE& lhs,
                                      const VectorArraySSE& rhs)
{
	const VectorSSE x = lhs[0] * rhs[0];
	const VectorSSE y = lhs[1] * rhs[1];
	const VectorSSE z = lhs[2] * rhs[2];
	return VectorArraySSE(x, y, z);
}


inline const VectorArraySSE operator-(const VectorArraySSE& lhs,
                                      const VectorArraySSE& rhs)
{
	const VectorSSE x = lhs[0] - rhs[0];
	const VectorSSE y = lhs[1] - rhs[1];
	const VectorSSE z = lhs[2] - rhs[2];
	return VectorArraySSE(x, y, z);
}


inline const VectorArraySSE operator+(const VectorArraySSE& lhs,
                                      const VectorArraySSE& rhs)
{
	const VectorSSE x = lhs[0] + rhs[0];
	const VectorSSE y = lhs[1] + rhs[1];
	const VectorSSE z = lhs[2] + rhs[2];
	return VectorArraySSE(x, y, z);
}


inline VectorSSE dotProduct(const VectorArraySSE& a, const VectorArraySSE& b)
{
	const VectorArraySSE& c(a*b);

	return c[0] + c[1] + c[2];
}


inline VectorArraySSE::VectorArraySSE(const VectorSSE& a, const VectorSSE& b, const VectorSSE& c)
{
	v[0] = a;
	v[1] = b;
	v[2] = c;
}


inline VectorArraySSE::VectorArraySSE(float x, float y, float z)
{
	v[0] = VectorSSE(x);
	v[1] = VectorSSE(y);
	v[2] = VectorSSE(z);
}

/*
 * Converts an ArrayOfStructures in a StructureOfArrays.
 * Taken from
 *	Using SSE and SSE2: Misconceptions and Reality
 *	Alex Klimovitski
 *	Intel Developer UPDATE Magazine
 *	March 2001
 */
inline VectorArraySSE::VectorArraySSE(const Vector3D* vec3, size_t numRays)
{
	assert(numRays == 4);  // this code is expected to work with 4 vectors

	const Vector3D& a = vec3[0];
	const Vector3D& b = vec3[1];
	const Vector3D& c = vec3[2];
	const Vector3D& d = vec3[3];

	VectorSSE xy10(_mm_loadl_pi(ZERO_VEC.vec, (__m64*)a.getAddress(0)));
	VectorSSE zw10(_mm_loadl_pi(ZERO_VEC.vec, (__m64*)a.getAddress(2)));
	xy10 = _mm_loadh_pi(xy10.vec, (__m64*)b.getAddress(0));
	zw10 = _mm_loadh_pi(zw10.vec, (__m64*)b.getAddress(2));

	VectorSSE xy32(_mm_loadl_pi(ZERO_VEC.vec, (__m64*)c.getAddress(0)));
	VectorSSE zw32(_mm_loadl_pi(ZERO_VEC.vec, (__m64*)c.getAddress(2)));
	xy32 = _mm_loadh_pi(xy32.vec, (__m64*)d.getAddress(0));
	zw32 = _mm_loadh_pi(zw32.vec, (__m64*)d.getAddress(2));

	v[0] = VectorSSE(_mm_shuffle_ps(xy10.vec, xy32.vec, _MM_SHUFFLE(2,0,2,0)));
	v[1] = VectorSSE(_mm_shuffle_ps(xy10.vec, xy32.vec, _MM_SHUFFLE(3,1,3,1)));
	v[2] = VectorSSE(_mm_shuffle_ps(zw10.vec, zw32.vec, _MM_SHUFFLE(2,0,2,0)));
//	VectorSSE w(_mm_shuffle_ps(zw10.vec, zw32.vec, _MM_SHUFFLE(3,1,3,1)));
}


inline VectorArray3D convertToVector3D(const VectorArraySSE& va)
{
	const __m128 x0y0x1y1 = _mm_unpacklo_ps( va[0].vec, va[1].vec );
	const __m128 x2y2x3y3 = _mm_unpackhi_ps( va[0].vec, va[1].vec );

	VectorArray3D retval = {{
//		Vector3D( _mm_shuffle_ps( x0y0x1y1, z0z1z2z3, _MM_SHUFFLE( 0, 1, 0, 0 ) ) ),
		Vector3D( _mm_shuffle_ps( x0y0x1y1, va[2].vec, _MM_SHUFFLE( 0, 0, 1, 0 ) ) ),
//		Vector3D( _mm_shuffle_ps( x0y0x1y1, z0z1z2z3, _MM_SHUFFLE( 2, 3, 1, 1 ) ) ),
		Vector3D( _mm_shuffle_ps( x0y0x1y1, va[2].vec, _MM_SHUFFLE( 1, 1, 3, 2 ) ) ),
//		Vector3D( _mm_shuffle_ps( x2y2x3y3, z0z1z2z3, _MM_SHUFFLE( 0, 1, 2, 2 ) ) ),
		Vector3D( _mm_shuffle_ps( x2y2x3y3, va[2].vec, _MM_SHUFFLE( 2, 2, 1, 0 ) ) ),
//		Vector3D( _mm_shuffle_ps( x2y2x3y3, z0z1z2z3, _MM_SHUFFLE( 2, 3, 3, 3 ) ) )
		Vector3D( _mm_shuffle_ps( x2y2x3y3, va[2].vec, _MM_SHUFFLE( 3, 3, 3, 2 ) ) )
    }};

    return retval;
}


inline void VectorArraySSE::normalize()
{
	const VectorSSE dx_squared = v[0] * v[0];
	const VectorSSE dy_squared = v[1] * v[1];
	const VectorSSE dz_squared = v[2] * v[2];
	const VectorSSE length_squared = dx_squared + dy_squared + dz_squared;

	const VectorSSE approx(_mm_rsqrt_ps(length_squared.vec));
	const VectorSSE muls = length_squared * approx * approx;
	const VectorSSE rsqrt = (HALF_VEC * approx) * (THREE_VEC - muls);

	v[0] *= rsqrt;
	v[1] *= rsqrt;
	v[2] *= rsqrt;
}

}}

#endif
