#ifdef VECTORSSE_H
/**
 * VectorSSE.cpp
 * 	NOTE: to be used as an include-file so that everything
 *        in here is inline
 *
 * (c) 2007, 2008 Thomas Unterthiner
 *
 */

#include "VectorSSE.h"

#include <xmmintrin.h>
#include <emmintrin.h>
#include <pmmintrin.h>
#include <iostream>


namespace util
{
namespace sse
{


extern const VectorSSE ZERO_VEC;
extern const VectorSSE MINUS_ONE_VEC;


inline VectorSSE::VectorSSE()
{
	v[0] = 0.f;
	v[1] = 0.f;
	v[2] = 0.f;
	v[3] = 0.f;
}


inline VectorSSE::VectorSSE(float x, float y, float z, float w)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
	v[3] = w;
}


inline VectorSSE::VectorSSE(float x)
{
	v[0] = v[1] = v[2] = v[3] = x;
}


inline VectorSSE::VectorSSE(const VectorSSE& other)
{
	vec = other.vec;
}


inline VectorSSE& VectorSSE::operator+= (const VectorSSE& other)
{
	vec = _mm_add_ps(vec , other.vec);
	return *this;
}


inline VectorSSE& VectorSSE::operator-= (const VectorSSE& other)
{
	vec = _mm_sub_ps(vec , other.vec);
	return *this;
}


inline VectorSSE& VectorSSE::operator*= (const VectorSSE& other)
{
	vec = _mm_mul_ps(vec , other.vec);
	return *this;
}


inline VectorSSE& VectorSSE::operator&= (const VectorSSE& other)
{
	vec = _mm_and_ps(vec , other.vec);
	return *this;
}


inline VectorSSE& VectorSSE::operator|= (const VectorSSE& other)
{
	vec = _mm_or_ps(vec , other.vec);
	return *this;
}



inline bool operator==(const VectorSSE& lhs, const VectorSSE& rhs)
{
	__m128 r = _mm_cmpeq_ps(lhs.vec, rhs.vec);
	return _mm_movemask_ps(r) != 0;
}


inline bool operator!=(const VectorSSE& lhs, const VectorSSE& rhs)
{
	return !(lhs == rhs);
}


inline void VectorSSE::sqrt()
{
	vec = _mm_sqrt_ps(vec);
}


inline bool VectorSSE::isZero(size_t idx) const
{
	return i[idx] == 0;
}


inline bool VectorSSE::isZero() const
{
/*
	return !(reinterpret_cast<int>(v[0]) | reinterpret_cast<int>(v[1])
		| reinterpret_cast<int>(v[2]) | reinterpret_cast<int>(v[3]));
*/
	return !(i[0] | i[1] | i[2] | i[3]);
}

inline const VectorSSE operator-(const VectorSSE& v)
{
	const VectorSSE retval(_mm_mul_ps(v.vec, MINUS_ONE_VEC));
	return retval;
}


inline const VectorSSE operator+(const VectorSSE& lhs, const VectorSSE& rhs)
{
	VectorSSE r(lhs);
	r += rhs;
	return r;
}


inline const VectorSSE operator-(const VectorSSE& lhs, const VectorSSE& rhs)
{
	VectorSSE r(lhs);
	r -= rhs;
	return r;
}


inline const VectorSSE operator*(const VectorSSE& lhs, const VectorSSE& rhs)
{
	VectorSSE r(lhs);
	r *= rhs;
	return r;
}


inline const VectorSSE operator/(const VectorSSE& lhs, const VectorSSE& rhs)
{
	__m128 r = _mm_div_ps(lhs.vec , rhs.vec);
	return VectorSSE(r);
}


inline const VectorSSE operator&(const VectorSSE& lhs, const VectorSSE& rhs)
{
	VectorSSE r(lhs);
	r &= rhs;
	return r;
}


inline const VectorSSE operator|(const VectorSSE& lhs, const VectorSSE& rhs)
{
	VectorSSE r(lhs);
	r |= rhs;
	return r;
}


inline const VectorSSE compareNeq(const VectorSSE& lhs, const VectorSSE& rhs)
{
	__m128 r = _mm_cmpneq_ps(lhs.vec, rhs.vec);
	return VectorSSE(r);
}

inline const VectorSSE compareGt(const VectorSSE& lhs, const VectorSSE& rhs)
{
	return VectorSSE(_mm_cmpgt_ps(lhs, rhs));
}


inline const VectorSSE compareLt(const VectorSSE& lhs, const VectorSSE& rhs)
{
	return VectorSSE(_mm_cmplt_ps(lhs, rhs));
}


inline const VectorSSE min(const VectorSSE& a, const VectorSSE& b)
{
	return VectorSSE(_mm_min_ps(a, b));
}

inline std::ostream& operator<< (std::ostream& stream, const VectorSSE& v)
{
	stream << '(' << v[0] << ", " << v[1] << ", " << v[2] << ", " << v[3] << ')';
	return stream;
}

}}

#endif
